from __future__ import annotations

import abc
import contextlib
import copy
import datetime
import json
import logging
import traceback

import numpy
import numpy.typing
import os.path
import pathlib
import shutil
import threading
import typing
import uuid

from nion.swift.model import DataItem
from nion.swift.model import HDF5Handler
from nion.swift.model import Migration
from nion.swift.model import Model
from nion.swift.model import NDataHandler
from nion.swift.model import Persistence
from nion.swift.model import StorageHandler
from nion.swift.model import Utility
from nion.utils import Event


# define the versions that get stored in the JSON files
PROFILE_VERSION = 2
PROJECT_VERSION = 3
PROJECT_VERSION_0_14 = 2


PersistentDictType = typing.Dict[str, typing.Any]
_NDArray = numpy.typing.NDArray[typing.Any]
_CreateStorageHandlerFn = typing.Type[StorageHandler.StorageHandler]


class ReaderInfo:
    def __init__(self,
                 properties: PersistentDictType,
                 changed_ref: typing.List[bool],
                 large_format: bool,
                 storage_handler: StorageHandler.StorageHandler,
                 identifier: str) -> None:
        self.properties = properties
        self.changed_ref = changed_ref
        self.large_format = large_format
        self.storage_handler = storage_handler
        self.identifier = identifier


class DataItemStorageAdapter:
    """Persistent storage for writing data item properties, relationships, and data to its storage handler."""

    def __init__(self, storage_handler: StorageHandler.StorageHandler, properties: PersistentDictType) -> None:
        self.__storage_handler = storage_handler
        self.__properties = properties

    def close(self) -> None:
        if self.__storage_handler:
            self.__storage_handler.close()
            self.__storage_handler = typing.cast(StorageHandler.StorageHandler, None)

    @property
    def properties(self) -> PersistentDictType:
        return self.__properties

    @property
    def storage_handler(self) -> StorageHandler.StorageHandler:
        return self.__storage_handler

    def rewrite_item(self, item: Persistence.PersistentObject) -> None:
        file_datetime = getattr(item, "created_local")
        self.__storage_handler.write_properties(Migration.transform_from_latest(copy.deepcopy(self.__properties)), file_datetime)

    def update_data(self, item: Persistence.PersistentObject, data: typing.Optional[_NDArray]) -> None:
        file_datetime = getattr(item, "created_local")
        if data is not None:
            self.__storage_handler.write_data(data, file_datetime)

    def reserve_data(self, item: Persistence.PersistentObject, data_shape: typing.Tuple[int, ...], data_dtype: numpy.typing.DTypeLike) -> None:
        file_datetime = getattr(item, "created_local")
        self.__storage_handler.reserve_data(data_shape, data_dtype, file_datetime)

    def load_data(self, item: Persistence.PersistentObject) -> typing.Optional[_NDArray]:
        return self.__storage_handler.read_data()


class MigrationReader(typing.Protocol):

    def get_storage_properties(self) -> PersistentDictType: ...

    def _get_migration_stages(self) -> typing.Sequence[ProjectStorageSystemMigrationStage]: ...

    def _find_data_items(self, migration_stage: ProjectStorageSystemMigrationStage) -> typing.Sequence[StorageHandler.StorageHandler]: ...

    def _is_storage_handler_large_format(self, storage_handler: StorageHandler.StorageHandler) -> bool: ...

    def _read_library_properties(self, migration_stage: ProjectStorageSystemMigrationStage) -> PersistentDictType: ...


class MigrationWriter(typing.Protocol):

    def _migrate_data_item(self, reader_info: ReaderInfo, index: int, count: int) -> typing.Optional[ReaderInfo]: ...

    def _migrate_library_properties(self, library_properties: PersistentDictType, reader_info_list: typing.List[ReaderInfo]) -> None: ...


def migrate_to_latest(source_project_storage_system: MigrationReader,
                      target_project_storage_system: MigrationWriter) -> None:
    """Migrate the library data in source to target, upgrading them in the process.

    If target is None, then migration is done in place.
    """
    library_properties = None
    data_item_uuids = set()
    reader_info_list = list()
    library_updates: typing.Dict[uuid.UUID, typing.Dict[str, typing.Any]] = dict()
    deletions = list()

    # iterate through migration stages from newest to oldest, reading data items, updating them to the latest
    # version, and copying them to the new library. migration stages are the high level directories representing
    # different library versions up to 13. after version 13, files are stored in project files which have their own
    # versioning. run newest to oldest so that deletions in newer libraries won't be migrated; nor will data items
    # in older projects with the same uuid in a newer project.

    for migration_stage in source_project_storage_system._get_migration_stages():

        # find all data items for the given migration stage and return a list of storage handlers.
        # examples of storage handlers are NDataHandler and HDF5Handler. these give low level access to the file.
        # every storage handler must be closed.
        storage_handlers = source_project_storage_system._find_data_items(migration_stage)

        # next, construct a list of ReaderInfo objects. ReaderInfo stores the properties portion of the data item,
        # whether it has been changed during migration, whether it is a large format file, its storage handler,
        # and an identifier key. this loop skips files that cannot be read but prints an error message.
        preliminary_reader_info_list: typing.List[ReaderInfo] = list()
        for storage_handler in storage_handlers:
            try:
                large_format = source_project_storage_system._is_storage_handler_large_format(storage_handler)
                storage_handler_properties = storage_handler.read_properties()
                assert storage_handler_properties is not None
                properties = Migration.transform_to_latest(storage_handler_properties)
                reader_info = ReaderInfo(properties, [False], large_format, storage_handler, storage_handler.reference)
                preliminary_reader_info_list.append(reader_info)
            except Exception:
                storage_handler.close()
                logging.debug("Error reading %s", storage_handler.reference)
                import traceback
                traceback.print_exc()
                traceback.print_stack()
            storage_handler.prepare_move()

        # now read the library properties which contains the data item deletions. data item deletions exist to
        # facilitate switching between library versions. if the user deletes an item in a newer library, that item
        # is marked as deleted so that if migration is performed again, that deleted item will not be re-migrated.
        new_library_properties = source_project_storage_system._read_library_properties(migration_stage)
        for deletion in copy.deepcopy(new_library_properties.get("data_item_deletions", list())):
            if deletion not in deletions:
                deletions.append(deletion)

        # set library properties to one from the first/newest migration stage encountered with library properties.
        if library_properties is None:
            library_properties = copy.deepcopy(new_library_properties)

        # next, for each item in the list of ReaderInfo objects, migrate it to the latest version. doing this may
        # produce additional library updates in preliminary_library_updates. these are changes to the library that
        # must be made in order to move information that at one point was stored in the data item files into the
        # library. an example is a computation, which was originally stored in the data item file itself.
        preliminary_library_updates: typing.Dict[uuid.UUID, PersistentDictType] = dict()
        Migration.migrate_to_latest(preliminary_reader_info_list, preliminary_library_updates)

        # finally, for each item in the preliminary_reader_info_list, confirm that it is the latest version and then
        # check whether it has a unique UUID that hasn't been deleted, and, if so, try to copy the data item to its
        # new location. if successful, mark the data item as having been added to the new library and add any
        # preliminary library updates to the library updates list to be applied later.
        count = len(preliminary_reader_info_list)
        for index, reader_info in enumerate(preliminary_reader_info_list):
            properties = reader_info.properties
            try:
                version = properties.get("version", 0)
                if version == DataItem.DataItem.writer_version:
                    data_item_uuid = uuid.UUID(properties["uuid"])
                    if data_item_uuid not in data_item_uuids:
                        if not str(data_item_uuid) in deletions:
                            new_reader_info = target_project_storage_system._migrate_data_item(reader_info, index, count)
                            if new_reader_info:
                                reader_info_list.append(new_reader_info)
                                data_item_uuids.add(data_item_uuid)
                                library_update = preliminary_library_updates.get(data_item_uuid)
                                if library_update:
                                    library_updates[data_item_uuid] = library_update
            except Exception:
                logging.debug(f"Error reading {reader_info.storage_handler.reference}")
                import traceback
                traceback.print_exc()
                traceback.print_stack()

        for storage_handler in storage_handlers:
            storage_handler.close()

    assert len(reader_info_list) == len(data_item_uuids)

    assert library_properties is not None

    # for each data item represented by a ReaderInfo object, apply its library updates. this will include
    # connections, computations, and display items. for instance, before version 13, the data item and display item
    # were both stored in the data item file; this migrates the display portion to the library properties.
    for reader_info in reader_info_list:
        properties = reader_info.properties
        properties = Utility.clean_dict(copy.deepcopy(properties) if properties else dict())
        version = properties.get("version", 0)
        if version == DataItem.DataItem.writer_version:
            data_item_uuid = uuid.UUID(typing.cast(str, properties.get("uuid", str(uuid.uuid4()))))
            library_update = library_updates.get(data_item_uuid, dict())
            assert library_update is not None
            library_properties.setdefault("connections", list()).extend(library_update.get("connections", list()))
            library_properties.setdefault("computations", list()).extend(library_update.get("computations", list()))
            library_properties.setdefault("display_items", list()).extend(library_update.get("display_items", list()))

    connections_list = library_properties.get("connections", list())
    assert len(connections_list) == len({connection.get("uuid") for connection in connections_list})

    computations_list = library_properties.get("computations", list())
    assert len(computations_list) == len({computation.get("uuid") for computation in computations_list})

    # migrate the library properties
    Migration.migrate_library_to_latest(library_properties)

    # TODO: add consistency checks: no duplicated items [by uuid] such as connections or computations or data items

    assert library_properties["version"] == PROJECT_VERSION

    # propagate the UUID
    source_library_properties = source_project_storage_system.get_storage_properties()
    if source_library_properties and "uuid" in source_library_properties:
        library_properties["uuid"] = source_library_properties["uuid"]

    target_project_storage_system._migrate_library_properties(library_properties, reader_info_list)


class PersistentStorageSystem(Persistence.PersistentStorageInterface):
    """Abstract base class for persistent storage which implements the persistent storage interface.

    Subclasses must implement _read_properties and _write_properties to read/write to persistent storage.

    The `load_properties` method must be called after instantiating the subclass.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__identifier = str(uuid.uuid4())  # only used when subclasses do not override
        self.__properties: PersistentDictType = dict()
        self.__properties_lock = threading.RLock()
        self.__write_delay_counts: typing.Dict[Persistence.PersistentObject, int] = dict()
        self.__write_delay_count = 0

    def close(self) -> None:
        pass

    def reset(self) -> None:
        pass

    def get_identifier(self) -> str:
        return self.__identifier

    @abc.abstractmethod
    def _write_properties(self) -> None:
        """Write internal properties, retrieved using _get_properties, to persistent storage."""
        ...

    @abc.abstractmethod
    def _read_properties(self) -> PersistentDictType:
        """Read internal properties from persistent storage."""
        ...

    def __set_persistent_storage(self, item: Persistence.PersistentObject, persistent_dict: typing.Optional[Persistence.PersistentDictType], persistent_storage: typing.Optional[Persistence.PersistentStorageInterface]) -> None:
        persistent_storage = typing.cast(typing.Optional[PersistentStorageSystem], persistent_storage)
        if persistent_storage:
            persistent_storage._unregister_persistent_dict(item)
        item.persistent_storage = persistent_storage
        if persistent_storage:
            persistent_storage._register_persistent_dict(item, persistent_dict)
        for key in item.item_names:
            component_item = item.get_item(key)
            if component_item:
                d = item.persistent_storage._get_item_persistent_dict(item, key) if item.persistent_storage else None
                self.__set_persistent_storage(component_item, d if persistent_dict is not None else None, item.persistent_storage if persistent_dict is not None else None)
        for key in item.relationship_names:
            for index, relationship_item in enumerate(item.get_relationship_items(key)):
                d = item.persistent_storage._get_relationship_persistent_dict(item, relationship_item, key, index) if item.persistent_storage else None
                self.__set_persistent_storage(relationship_item, d if persistent_dict is not None else None, item.persistent_storage if persistent_dict is not None else None)

    def set_root_item(self, item: Persistence.PersistentObject) -> None:
        """Set the storage system for this item."""
        self.__set_persistent_storage(item, self.get_storage_properties(), self)

    def unload_item(self, item: Persistence.PersistentObject) -> None:
        self._unregister_persistent_dict(item)

    def load_properties(self) -> None:
        """Read properties and store them in internal storage. Should be called immediately after instantiation."""
        with self.__properties_lock:
            self.__properties = self._read_properties()

    def read_project_properties(self) -> typing.Tuple[PersistentDictType, typing.Sequence[Persistence.ReaderError]]:
        # dummy implementation for now until this is combined with load_properties.
        # subclass may override.
        return dict(), list()

    def get_storage_properties(self) -> PersistentDictType:
        """Return the internal properties. Callers should not modify; it is ok to not return a copy."""
        return self.__properties

    def migrate_to_latest(self) -> None:
        pass

    def _register_persistent_dict(self, item: Persistence.PersistentObject, persistent_dict: typing.Optional[PersistentDictType]) -> None:
        setattr(item, "__persistent_dict", persistent_dict)

    def _unregister_persistent_dict(self, item: Persistence.PersistentObject) -> None:
        setattr(item, "__persistent_dict", None)

    def _get_persistent_dict(self, item: Persistence.PersistentObject) -> typing.Optional[PersistentDictType]:
        return getattr(item, "__persistent_dict", None)

    def _get_item_persistent_dict(self, container: Persistence.PersistentObject, key: str) -> typing.Optional[PersistentDictType]:
        d = getattr(container, "__persistent_dict", None)
        return d[key] if d is not None else None

    def _get_relationship_persistent_dict(self, container: Persistence.PersistentObject, item: Persistence.PersistentObject, key: str, index: int) -> typing.Optional[PersistentDictType]:
        d = getattr(container, "__persistent_dict", None)
        return d[key][index] if d is not None else None

    def _get_relationship_persistent_dict_by_uuid(self, container: Persistence.PersistentObject, item: Persistence.PersistentObject, key: str) -> typing.Optional[PersistentDictType]:
        d = getattr(container, "__persistent_dict", None)
        if d is not None:
            item_uuid = str(item.uuid)
            for item_d in d.get(key, list()):
                if item_d.get("uuid") == item_uuid:  # a little dangerous, comparing the uuid str's, significantly faster
                    return typing.cast(PersistentDictType, item_d)
        return None

    def __write_properties_if_not_delayed(self, item: typing.Optional[Persistence.PersistentObject]) -> None:
        if not item or self.__write_delay_counts.get(item, 0) == 0:
            self._write_item_properties(item)

    def _write_item_properties(self, item: typing.Optional[Persistence.PersistentObject]) -> None:
        persistent_object_parent = item.persistent_object_parent if item else None
        if not persistent_object_parent:
            if self.__write_delay_count == 0:
                self._write_properties()
        else:
            self.__write_properties_if_not_delayed(persistent_object_parent.parent)

    def get_properties(self, item: Persistence.PersistentObject) -> typing.Optional[PersistentDictType]:
        return self._get_persistent_dict(item)

    def __update_modified_and_get_storage_dict(self, item: Persistence.PersistentObject) -> PersistentDictType:
        # update modified time on object and all parent objects in internal storage
        storage_dict = self._get_persistent_dict(item)
        assert storage_dict is not None
        with self.__properties_lock:
            storage_dict["modified"] = item.modified.isoformat()
        persistent_object_parent = item.persistent_object_parent
        parent = persistent_object_parent.parent if persistent_object_parent else None
        if parent:
            self.__update_modified_and_get_storage_dict(parent)
        return storage_dict

    def insert_relationship_item(self, parent: Persistence.PersistentObject, name: str, before_index: int, item: Persistence.PersistentObject) -> None:
        # insert item in internal storage
        self.__set_persistent_storage(item, item.write_to_dict(), self)
        self._insert_item(parent, name, before_index, item)

    def remove_relationship_item(self, parent: Persistence.PersistentObject, name: str, index: int, item: Persistence.PersistentObject) -> None:
        self._remove_item(parent, name, index, item)
        self.__set_persistent_storage(item, None, None)

    def load_relationship_item(self, parent: Persistence.PersistentObject, name: str, before_index: int, item: Persistence.PersistentObject) -> None:
        d = self._get_relationship_persistent_dict_by_uuid(parent, item, name) or dict()
        self.__set_persistent_storage(item, d, self)

    def unload_relationship_item(self, parent: Persistence.PersistentObject, item: Persistence.PersistentObject) -> None:
        self.__set_persistent_storage(item, None, None)

    def load_component_item(self, parent: Persistence.PersistentObject, name: str, item: Persistence.PersistentObject) -> None:
        d = self._get_item_persistent_dict(parent, name) or dict()
        self.__set_persistent_storage(item, d, self)

    def _insert_item(self, parent: Persistence.PersistentObject, name: str, before_index: int, item: Persistence.PersistentObject) -> None:
        storage_dict = self.__update_modified_and_get_storage_dict(parent)
        with self.__properties_lock:
            item_list = storage_dict.setdefault(name, list())
            item_list.insert(before_index, self._get_persistent_dict(item))
        self.__write_properties_if_not_delayed(parent)

    def _remove_item(self, parent: Persistence.PersistentObject, name: str, index: int, item: Persistence.PersistentObject) -> None:
        # remove item from internal storage
        storage_dict = self.__update_modified_and_get_storage_dict(parent)
        with self.__properties_lock:
            item_list = storage_dict[name]
            del item_list[index]
        self.__write_properties_if_not_delayed(parent)

    def set_component_item(self, parent: Persistence.PersistentObject, name: str, item: typing.Optional[Persistence.PersistentObject]) -> None:
        storage_dict = self.__update_modified_and_get_storage_dict(parent)
        if item:
            # set the item and update its persistent context
            with self.__properties_lock:
                self.__set_persistent_storage(item, item.write_to_dict(), self)
                storage_dict[name] = self._get_persistent_dict(item)
        else:
            # clear the item
            with self.__properties_lock:
                storage_dict.pop(name, None)
                if item:
                    self.__set_persistent_storage(item, None, None)
        self.__write_properties_if_not_delayed(parent)

    def set_property(self, object: Persistence.PersistentObject, name: str, value: typing.Any, delayed: bool = False) -> None:
        # set property in internal storage. if the delayed flag is set, it will not trigger a write to disk.
        storage_dict = self.__update_modified_and_get_storage_dict(object)
        with self.__properties_lock:
            storage_dict[name] = value
        if not delayed:
            self.__write_properties_if_not_delayed(object)

    def clear_property(self, object: Persistence.PersistentObject, name: str) -> None:
        # clear property in internal storage
        storage_dict = self.__update_modified_and_get_storage_dict(object)
        with self.__properties_lock:
            storage_dict.pop(name, None)
        self.__write_properties_if_not_delayed(object)

    def get_storage_property(self, item: Persistence.PersistentObject, name: str) -> typing.Optional[str]:
        return None

    def read_external_data(self, item: Persistence.PersistentObject, name: str) -> typing.Any:
        return None

    def write_external_data(self, item: Persistence.PersistentObject, name: str, value: _NDArray) -> None:
        pass

    def reserve_external_data(self, item: Persistence.PersistentObject, name: str, data_shape: typing.Tuple[int, ...],
                              data_dtype: numpy.typing.DTypeLike) -> None:
        pass

    def enter_write_delay(self, object: Persistence.PersistentObject) -> None:
        count = self.__write_delay_counts.setdefault(object, 0)
        self.__write_delay_counts[object] = count + 1

    def exit_write_delay(self, object: Persistence.PersistentObject) -> None:
        count = self.__write_delay_counts.get(object, 1)
        count -= 1
        if count == 0:
            self.__write_delay_counts.pop(object)
        else:
            self.__write_delay_counts[object] = count

    def is_write_delayed(self, item: Persistence.PersistentObject) -> bool:
        return self.__write_delay_counts.get(item, 0) > 0

    def rewrite_item(self, item: Persistence.PersistentObject) -> None:
        self.__write_properties_if_not_delayed(item)

    def restore_item(self, data_item_uuid: uuid.UUID) -> typing.Optional[PersistentDictType]:
        raise NotImplementedError()

    def prune(self) -> None:
        pass

    def enter_transaction(self) -> None:
        self.__write_delay_count += 1

    def exit_transaction(self) -> None:
        self.__write_delay_count -= 1
        if self.__write_delay_count == 0:
            self.__write_properties_if_not_delayed(None)

    def _get_persistence_write_count(self, item: Persistence.PersistentObject) -> typing.Optional[int]:
        return None


class FilePersistentStorageSystem(PersistentStorageSystem):
    """File based persistent storage system."""

    def __init__(self, path: pathlib.Path) -> None:
        self.__path = path
        super().__init__()

    def close(self) -> None:
        pass

    @property
    def path(self) -> pathlib.Path:
        return self.__path

    def _read_properties(self) -> PersistentDictType:
        properties = dict()
        if self.__path and self.__path.exists():
            try:
                with self.__path.open("r") as fp:
                    properties = json.load(fp)
            except Exception:
                os.replace(self.__path, self.__path.with_suffix(".bak"))
        return properties

    def _write_properties(self) -> None:
        if self.__path:
            with Utility.AtomicFileWriter(self.__path) as fp:
                properties = Utility.clean_dict(self.get_storage_properties())
                json.dump(properties, fp)


class MemoryPersistentStorageSystem(PersistentStorageSystem):
    """File based persistent storage system. Useful for testing."""

    def __init__(self, *, library_properties: typing.Optional[PersistentDictType] = None) -> None:
        self.__library_properties = library_properties if library_properties is not None else dict()
        super().__init__()

    def close(self) -> None:
        pass

    def _read_properties(self) -> PersistentDictType:
        return copy.deepcopy(self.__library_properties)

    def _write_properties(self) -> None:
        self.__library_properties.clear()
        self.__library_properties.update(self.get_storage_properties())

    def set_library_properties(self, library_properties: PersistentDictType) -> None:
        self.__library_properties.clear()
        self.__library_properties.update(library_properties)
        self.load_properties()


class ProjectStorageSystemMigrationStage:
    pass


class ProjectStorageSystem(PersistentStorageSystem):
    """Persistent storage system to provide special handling of data items."""

    def __init__(self) -> None:
        super().__init__()
        self.__storage_adapter_map: typing.Dict[uuid.UUID, DataItemStorageAdapter] = dict()

    def close(self) -> None:
        for storage_adapter in self.__storage_adapter_map.values():
            storage_adapter.close()
        self.__storage_adapter_map.clear()

    @abc.abstractmethod
    def _make_storage_handler(self, data_item: DataItem.DataItem, file_handler: typing.Optional[StorageHandler.StorageHandlerFactoryLike] = None) -> StorageHandler.StorageHandler: ...

    @abc.abstractmethod
    def _find_storage_handlers(self) -> typing.Sequence[StorageHandler.StorageHandler]: ...

    @abc.abstractmethod
    def _remove_storage_handler(self, storage_handler: StorageHandler.StorageHandler, *, safe: bool = False) -> None: ...

    @abc.abstractmethod
    def _restore_item(self, data_item_uuid: uuid.UUID) -> typing.Optional[PersistentDictType]: ...

    @abc.abstractmethod
    def _prune(self) -> None: ...

    # for migration

    @abc.abstractmethod
    def _get_migration_stages(self) -> typing.Sequence[ProjectStorageSystemMigrationStage]: ...

    @abc.abstractmethod
    def _find_data_items(self, migration_stage: ProjectStorageSystemMigrationStage) -> typing.Sequence[StorageHandler.StorageHandler]: ...

    @abc.abstractmethod
    def _is_storage_handler_large_format(self, storage_handler: StorageHandler.StorageHandler) -> bool: ...

    @abc.abstractmethod
    def _migrate_data_item(self, reader_info: ReaderInfo, index: int, count: int) -> typing.Optional[ReaderInfo]: ...

    @abc.abstractmethod
    def _migrate_library_properties(self, library_properties: PersistentDictType, reader_info_list: typing.List[ReaderInfo]) -> None: ...

    @abc.abstractmethod
    def _read_library_properties(self, migration_stage: ProjectStorageSystemMigrationStage) -> PersistentDictType: ...

    #

    @property
    def _data_properties_map(self) -> typing.Dict[uuid.UUID, DataItemStorageAdapter]:
        return self.__storage_adapter_map

    def reset(self) -> None:
        self.__storage_adapter_map = dict()

    def _get_persistence_write_count(self, item: Persistence.PersistentObject) -> typing.Optional[int]:
        return getattr(self._data_properties_map[item.uuid].storage_handler, "_write_count", None)

    def _get_relationship_persistent_dict(self, container: Persistence.PersistentObject, item: Persistence.PersistentObject, key: str, index: int) -> typing.Optional[PersistentDictType]:
        if key == "data_items":
            return self._data_properties_map[item.uuid].properties
        else:
            return super()._get_relationship_persistent_dict(container, item, key, index)

    def _get_relationship_persistent_dict_by_uuid(self, container: Persistence.PersistentObject, item: Persistence.PersistentObject, key: str) -> typing.Optional[PersistentDictType]:
        if key == "data_items":
            return self._data_properties_map[item.uuid].properties
        else:
            return super()._get_relationship_persistent_dict_by_uuid(container, item, key)

    def get_persistent_dict(self, name: str, item_uuid: uuid.UUID) -> PersistentDictType:
        if name == "data_items":
            return self._data_properties_map[item_uuid].properties
        for item_d in self.get_storage_properties()[name]:
            if uuid.UUID(item_d["uuid"]) == item_uuid:
                return typing.cast(PersistentDictType, item_d)
        assert False

    def read_project_properties(self) -> typing.Tuple[PersistentDictType, typing.Sequence[Persistence.ReaderError]]:
        """Read data items from the data reference handler and return as a dict.

        The dict may contain keys for data_items, display_items, data_structures, connections, and computations.
        """
        storage_handlers = self._find_storage_handlers()

        reader_info_list = list()
        reader_error_list = list()
        for storage_handler in storage_handlers:
            try:
                large_format = self._is_storage_handler_large_format(storage_handler)
                storage_handler_properties = storage_handler.read_properties()
                storage_handler.prepare_move()
                assert storage_handler_properties is not None
                properties = Migration.transform_to_latest(storage_handler_properties)
                reader_info = ReaderInfo(properties, [False], large_format, storage_handler, storage_handler.reference)
                reader_info_list.append(reader_info)
            except Exception as e:
                reader_error_list.append(Persistence.ReaderError(storage_handler.reference, e, traceback.extract_stack()))

        # to allow later writing back to storage, associate the data items with their storage adapters
        for reader_info in reader_info_list:
            storage_handler = reader_info.storage_handler
            properties = reader_info.properties
            data_item_uuid = uuid.UUID(properties["uuid"])
            storage_adapter = DataItemStorageAdapter(storage_handler, properties)
            old_storage_adapter = self.__storage_adapter_map.pop(data_item_uuid, None)
            if old_storage_adapter:
                old_storage_adapter.close()
            self.__storage_adapter_map[data_item_uuid] = storage_adapter

        properties_copy = self._read_properties()

        # ensure unique connections
        connections_list = properties_copy.get("connections", list())
        assert len(connections_list) == len({connection.get("uuid") for connection in connections_list})

        # ensure unique computations
        computations_list = properties_copy.get("computations", list())
        assert len(computations_list) == len({computation.get("uuid") for computation in computations_list})

        for reader_info in reader_info_list:
            data_item_properties = reader_info.properties if reader_info.properties else dict()
            if data_item_properties.get("version", 0) == DataItem.DataItem.writer_version:
                data_item_properties["__large_format"] = reader_info.large_format
                properties_copy.setdefault("data_items", list()).append(data_item_properties)

        def data_item_created(data_item_properties: PersistentDictType) -> str:
            # created is a utc timestamp
            earliest_datetime = datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc).isoformat()
            return typing.cast(str, data_item_properties.get("created", earliest_datetime))

        data_items_copy = sorted(properties_copy.get("data_items", list()), key=data_item_created)
        if len(data_items_copy) > 0:
            properties_copy["data_items"] = data_items_copy

        return Model.transform_forward(properties_copy), reader_error_list

    # override
    def _write_item_properties(self, item: typing.Optional[Persistence.PersistentObject]) -> None:
        if item and isinstance(item, DataItem.DataItem):
            self.__rewrite_data_item_properties(item)
        else:
            super()._write_item_properties(item)

    # override
    def _insert_item(self, parent: Persistence.PersistentObject, name: str, before_index: int, item: Persistence.PersistentObject) -> None:
        if isinstance(item, DataItem.DataItem):
            item_uuid = item.uuid
            storage_handler = self._make_storage_handler(item)
            assert item_uuid not in self.__storage_adapter_map
            persistent_dict = self._get_persistent_dict(item)
            assert persistent_dict is not None
            storage_adapter = DataItemStorageAdapter(storage_handler, persistent_dict)
            self.__storage_adapter_map[item_uuid] = storage_adapter
        else:
            super()._insert_item(parent, name, before_index, item)

    # override
    def _remove_item(self, parent: Persistence.PersistentObject, name: str, index: int, item: Persistence.PersistentObject) -> None:
        if isinstance(item, DataItem.DataItem):
            assert item.uuid in self.__storage_adapter_map
            storage = self.__storage_adapter_map.get(item.uuid)
            assert storage
            self._remove_storage_handler(storage.storage_handler, safe=True)
            self.__storage_adapter_map.pop(item.uuid).close()
        else:
            super()._remove_item(parent, name, index, item)

    # override
    def get_storage_property(self, item: Persistence.PersistentObject, name: str) -> typing.Optional[str]:
        if isinstance(item, DataItem.DataItem):
            return self.__get_data_item_property(item, name)
        return super().get_storage_property(item, name)

    # override
    def read_external_data(self, item: Persistence.PersistentObject, name: str) -> typing.Any:
        if isinstance(item, DataItem.DataItem) and name == "data":
            return self.__read_data_item_data(item)
        return super().read_external_data(item, name)

    # override
    def write_external_data(self, item: Persistence.PersistentObject, name: str, value: _NDArray) -> None:
        if isinstance(item, DataItem.DataItem) and name == "data":
            self.__write_data_item_data(item, value)
        else:
            super().write_external_data(item, name, value)

    # override
    def reserve_external_data(self, item: Persistence.PersistentObject, name: str, data_shape: typing.Tuple[int, ...],
                              data_dtype: numpy.typing.DTypeLike) -> None:
        if isinstance(item, DataItem.DataItem) and name == "data":
            self.__reserve_data_item_data(item, data_shape, numpy.dtype(data_dtype))
        else:
            super().reserve_external_data(item, name, data_shape, data_dtype)

    # override
    def rewrite_item(self, item: Persistence.PersistentObject) -> None:
        if isinstance(item, DataItem.DataItem):
            self.__rewrite_data_item_properties(item)
        else:
            super().rewrite_item(item)

    def restore_item(self, data_item_uuid: uuid.UUID) -> typing.Optional[PersistentDictType]:
        return self.__restore_item(data_item_uuid)

    def prune(self) -> None:
        self._prune()

    def find_data_items(self) -> typing.Sequence[StorageHandler.StorageHandler]:
        return self._find_storage_handlers()

    def migrate_to_latest(self) -> None:
        migrate_to_latest(self, self)

    def __get_data_item_properties(self, data_item: DataItem.DataItem) -> PersistentDictType:
        storage_adapter = self.__storage_adapter_map.get(data_item.uuid)
        assert storage_adapter
        return storage_adapter.properties

    def __get_data_item_property(self, data_item: DataItem.DataItem, name: str) -> typing.Optional[str]:
        if name == "file_path":
            storage = self.__storage_adapter_map.get(data_item.uuid)
            return storage.storage_handler.reference if storage else None
        return None

    def __read_data_item_data(self, data_item: DataItem.DataItem) -> typing.Optional[_NDArray]:
        storage_adapter = self.__storage_adapter_map.get(data_item.uuid)
        assert storage_adapter
        return storage_adapter.load_data(data_item)

    def __write_data_item_data(self, data_item: DataItem.DataItem, data: typing.Optional[_NDArray]) -> None:
        if not self.is_write_delayed(data_item):
            storage_adapter = self.__storage_adapter_map.get(data_item.uuid)
            assert storage_adapter
            storage_adapter.update_data(data_item, data)

    def __reserve_data_item_data(self, data_item: DataItem.DataItem, data_shape: typing.Tuple[int, ...], data_dtype: numpy.typing.DTypeLike) -> None:
        storage_adapter = self.__storage_adapter_map.get(data_item.uuid)
        assert storage_adapter
        storage_adapter.reserve_data(data_item, data_shape, data_dtype)

    def __rewrite_data_item_properties(self, data_item: DataItem.DataItem) -> None:
        if not self.is_write_delayed(data_item):
            storage_adapter = self.__storage_adapter_map.get(data_item.uuid)
            assert storage_adapter
            storage_adapter.rewrite_item(data_item)

    def __restore_item(self, data_item_uuid: uuid.UUID) -> typing.Optional[PersistentDictType]:
        return self._restore_item(data_item_uuid)


class FileProjectStorageSystemMigrationStage(ProjectStorageSystemMigrationStage):
    def __init__(self, library_path: pathlib.Path, library_folder: pathlib.Path) -> None:
        self.library_path = library_path
        self.library_folder = library_folder



class FileProjectStorageSystem(ProjectStorageSystem):

    _file_handler_factories: typing.List[StorageHandler.StorageHandlerFactoryLike] = [NDataHandler.NDataHandlerFactory(), HDF5Handler.HDF5HandlerFactory()]

    def __init__(self, project_path: pathlib.Path, project_data_path: typing.Optional[pathlib.Path] = None) -> None:
        super().__init__()
        self.__project_path = project_path
        self.__project_data_path = project_data_path

    def load_properties(self) -> None:
        # in order to be resilient to name changes, first make a list of folders in project_data_folders which
        # (1) can be constructed and (2) which exist. if none actually exist, see if one exists based on the
        # root name. then, if a folder exists, use it. otherwise, use the first one that can be constructed from
        # either the project_data_folders or the root name. this method is expected to supply a project_data_path
        # that can be written; the case where the project is created is where the folder will not exist in the
        # first place.
        super().load_properties()
        project_data_folder_paths = list()
        existing_project_data_folder_paths = list()
        for project_data_folder in self.get_storage_properties().get("project_data_folders", list()):
            project_data_folder_path = pathlib.Path(project_data_folder)
            if not project_data_folder_path.is_absolute():
                project_data_folder_path = self.__project_path.parent / project_data_folder_path
            project_data_folder_paths.append(project_data_folder_path)
            if project_data_folder_path.exists():
                existing_project_data_folder_paths.append(project_data_folder_path)
        if not existing_project_data_folder_paths:
            project_data_folder_path = self.__project_path.with_name(self.__project_path.stem + " Data")
            if not project_data_folder_path.is_absolute():
                project_data_folder_path = self.__project_path.parent / project_data_folder_path
            project_data_folder_paths.append(project_data_folder_path)
            if project_data_folder_path.exists():
                existing_project_data_folder_paths.append(project_data_folder_path)
        if existing_project_data_folder_paths:
            self.__project_data_path = existing_project_data_folder_paths[0]
        else:
            self.__project_data_path = project_data_folder_paths[0] if len(project_data_folder_paths) > 0 else self.__project_data_path

    @property
    def project_path(self) -> pathlib.Path:
        return self.__project_path

    def _read_properties(self) -> PersistentDictType:
        properties = dict()
        if self.__project_path and self.__project_path.exists():
            with self.__project_path.open("r") as fp:
                properties = json.load(fp)
        return properties

    def _write_properties(self) -> None:
        self.__write_properties_inner(Model.transform_backward(copy.deepcopy(self.get_storage_properties())))

    def __write_properties_inner(self, properties: PersistentDictType) -> None:
        if self.__project_path:
            # atomically overwrite
            with Utility.AtomicFileWriter(self.__project_path) as fp:
                properties = Utility.clean_dict(properties)
                project_data_paths = list()
                for project_data_path in [self.__project_data_path] if self.__project_data_path else []:
                    if project_data_path.parent == self.__project_path.parent:
                        project_data_path = project_data_path.relative_to(project_data_path.parent)
                    project_data_paths.append(project_data_path)
                project_uuid = uuid.uuid4()
                properties.setdefault("uuid", str(project_uuid))
                properties["project_data_folders"] = [str(project_data_path) for project_data_path in project_data_paths]
                json.dump(properties, fp)

    def get_identifier(self) -> str:
        return str(self.__project_path)

    def _make_storage_handler(self, data_item: DataItem.DataItem, file_handler_factory: typing.Optional[StorageHandler.StorageHandlerFactoryLike] = None) -> StorageHandler.StorageHandler:
        # if there are two handlers, first is small, second is large
        # if there is only one handler, it is used in all cases
        large_format = hasattr(data_item, "large_format") and data_item.large_format
        file_handler_factory = file_handler_factory if file_handler_factory else (self._file_handler_factories[-1] if large_format else self._file_handler_factories[0])
        assert self.__project_data_path is not None
        return file_handler_factory.make(self.__project_data_path / self.__get_base_path(data_item))

    def _find_storage_handlers(self) -> typing.Sequence[StorageHandler.StorageHandler]:
        return self.__find_storage_handlers(self.__project_data_path)

    def _is_storage_handler_large_format(self, storage_handler: StorageHandler.StorageHandler) -> bool:
        return isinstance(storage_handler, HDF5Handler.HDF5Handler)

    def _remove_storage_handler(self, storage_handler: StorageHandler.StorageHandler, *, safe: bool = False) -> None:
        assert self.__project_data_path is not None
        file_path = pathlib.Path(storage_handler.reference)
        file_name = file_path.parts[-1]
        trash_dir = self.__project_data_path / "trash"
        new_file_path = trash_dir / file_name
        storage_handler.prepare_move()  # moving files in the storage handler requires it to be closed.
        # TODO: move this functionality to the storage handler.
        if safe and not os.path.exists(new_file_path):
            trash_dir.mkdir(exist_ok=True)
            shutil.move(str(file_path), new_file_path)
        storage_handler.remove()

    def _restore_item(self, data_item_uuid: uuid.UUID) -> typing.Optional[PersistentDictType]:
        assert self.__project_data_path is not None
        data_item_uuid_str = str(data_item_uuid)
        trash_dir = self.__project_data_path / "trash"
        storage_handlers = self.__find_storage_handlers(trash_dir, skip_trash=False)
        try:
            for storage_handler in storage_handlers:
                storage_handler_properties = storage_handler.read_properties()
                assert storage_handler_properties is not None
                properties = Migration.transform_to_latest(storage_handler_properties)
                if properties.get("uuid", None) == data_item_uuid_str:
                    data_item = DataItem.DataItem(item_uuid=data_item_uuid)
                    with contextlib.closing(data_item):
                        data_item.begin_reading()
                        data_item.read_from_dict(properties)
                        data_item.finish_reading()
                        old_file_path = storage_handler.reference
                        new_file_path = storage_handler.factory.make_path(self.__project_data_path / self.__get_base_path(data_item))
                        if not os.path.exists(new_file_path):
                            os.makedirs(os.path.dirname(new_file_path), exist_ok=True)
                            shutil.move(old_file_path, new_file_path)
                        self._make_storage_handler(data_item).close()  # what's this line for?
                        properties["__large_format"] = isinstance(storage_handler, HDF5Handler.HDF5Handler)
                        return properties
        finally:
            for storage_handler in storage_handlers:
                storage_handler.close()
        return None

    def _prune(self) -> None:
        if self.__project_data_path:
            trash_dir = self.__project_data_path / "trash"
            for file_path in trash_dir.rglob("*"):
                # the date is not a reliable way of determining the age since a user may trash an old file. for now,
                # we just delete anything in the trash at startup. future version may have an index file for
                # tracking items in the trash. when items are again retained in the trash, update the disabled
                # test_delete_and_undelete_from_file_storage_system_restores_data_item_after_reload
                file_path.unlink()

    @property
    def _trash_dir(self) -> pathlib.Path:
        assert self.__project_data_path is not None
        return self.__project_data_path / "trash"

    def _migrate_data_item(self, reader_info: ReaderInfo, index: int, count: int) -> typing.Optional[ReaderInfo]:
        storage_handler = reader_info.storage_handler
        properties = reader_info.properties
        properties = Utility.clean_dict(copy.deepcopy(properties) if properties else dict())
        data_item_uuid = uuid.UUID(typing.cast(str, properties["uuid"]))
        old_data_item = DataItem.DataItem(item_uuid=data_item_uuid)
        with contextlib.closing(old_data_item):
            old_data_item.begin_reading()
            old_data_item.read_from_dict(properties)
            old_data_item.finish_reading()
            old_data_item_path = storage_handler.reference
            # ask the storage system for the file handler for the data item path
            file_handler_factory = self.__get_file_handler_factory_for_file(str(old_data_item_path))
            # ask the storage system to make a storage handler (an instance of a file handler) for the data item
            # this ensures that the storage handler (file format) is the same as before.
            with contextlib.closing(self._make_storage_handler(old_data_item, file_handler_factory)) as target_storage_handler:
                if target_storage_handler and storage_handler.reference != target_storage_handler.reference:
                    os.makedirs(os.path.dirname(target_storage_handler.reference), exist_ok=True)
                    target_storage_handler.prepare_move()
                    shutil.copyfile(storage_handler.reference, target_storage_handler.reference)
                    shutil.copystat(storage_handler.reference, target_storage_handler.reference)
                    target_storage_handler.write_properties(Migration.transform_from_latest(copy.deepcopy(properties)), datetime.datetime.now())
                    logging.getLogger("migration").info(f"Copying data item ({index + 1}/{count}) {data_item_uuid} to new library.")
                    return ReaderInfo(properties, [False], self._is_storage_handler_large_format(target_storage_handler),
                                      target_storage_handler, target_storage_handler.reference)
            logging.getLogger("migration").warning(f"Unable to copy data item {data_item_uuid} to new library.")
            return None

    def _migrate_library_properties(self, library_properties: PersistentDictType, reader_info_list: typing.List[ReaderInfo]) -> None:
        self.__write_properties_inner(library_properties)
        for reader_info in reader_info_list:
            data_item_properties = Utility.clean_dict(reader_info.properties if reader_info.properties else dict())
            if data_item_properties.get("version", 0) == DataItem.DataItem.writer_version:
                # file modified dates are stored as local timestamps
                earliest_datetime = datetime.datetime.fromtimestamp(0).isoformat()
                created = typing.cast(str, data_item_properties.get("created", earliest_datetime))
                file_datetime = DataItem.DatetimeToStringConverter().convert_back(created) or datetime.datetime.now()
                # storage handler has already been closed; writing properties MAY reopen it.
                # close it by "prepare for move".
                # this should be redesigned so that storage handler lifetime is well defined.
                # TODO: storage handler open/close is bad design.
                reader_info.storage_handler.write_properties(reader_info.properties, file_datetime)
                reader_info.storage_handler.prepare_move()

    @staticmethod
    def _get_migration_paths(library_path: pathlib.Path) -> typing.List[typing.Tuple[pathlib.Path, pathlib.Path]]:
        return [
            (library_path / "Nion Swift Library 13.nslib", library_path / "Nion Swift Data 13"),
            (library_path / "Nion Swift Library 12.nslib", library_path / "Nion Swift Data 12"),
            (library_path / "Nion Swift Workspace.nslib", library_path / "Nion Swift Data 11"),
            (library_path / "Nion Swift Workspace.nslib", library_path / "Nion Swift Data 10"),
            (library_path / "Nion Swift Workspace.nslib", library_path / "Nion Swift Data"),
        ]

    def _get_migration_stages(self) -> typing.Sequence[ProjectStorageSystemMigrationStage]:
        return list([FileProjectStorageSystemMigrationStage(p, f) for p, f in self._get_migration_paths(self.__project_path.parent)])

    def _read_library_properties(self, migration_stage: ProjectStorageSystemMigrationStage) -> PersistentDictType:
        migration_stage = typing.cast(FileProjectStorageSystemMigrationStage, migration_stage)
        properties = dict()
        project_path = migration_stage.library_path
        if project_path and os.path.exists(project_path):
            try:
                with project_path.open("r") as fp:
                    properties = json.load(fp)
            except Exception:
                os.replace(project_path, project_path.with_suffix(".bak"))
        return properties

    def _find_data_items(self, migration_stage: ProjectStorageSystemMigrationStage) -> typing.Sequence[StorageHandler.StorageHandler]:
        migration_stage = typing.cast(FileProjectStorageSystemMigrationStage, migration_stage)
        return self.__find_storage_handlers(migration_stage.library_folder)

    def __get_base_path(self, data_item: DataItem.DataItem) -> pathlib.Path:
        data_item_uuid = data_item.uuid
        created_local = data_item.created_local
        session_id = data_item.session_id
        # data_item_uuid.bytes.encode('base64').rstrip('=\n').replace('/', '_')
        # and back: data_item_uuid = uuid.UUID(bytes=(slug + '==').replace('_', '/').decode('base64'))
        # also:

        def encode(uuid_: uuid.UUID, alphabet: str) -> str:
            result = str()
            uuid_int = uuid_.int
            while uuid_int:
                uuid_int, digit = divmod(uuid_int, len(alphabet))
                result += alphabet[digit]
            return result

        path_components = created_local.strftime("%Y-%m-%d").split('-')
        session_id = session_id if session_id else created_local.strftime("%Y%m%d-000000")
        path_components.append(session_id)
        encoded_base_path = "data_" + encode(data_item_uuid, "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890")  # 25 character results
        path_components.append(encoded_base_path)
        return pathlib.Path(*path_components)

    def __get_file_handler_factory_for_file(self, path: str) -> typing.Optional[StorageHandler.StorageHandlerFactoryLike]:
        for file_handler_factory in self._file_handler_factories:
            if file_handler_factory.is_matching(path):
                return file_handler_factory
        return None

    def __find_storage_handlers(self, directory: typing.Optional[pathlib.Path], *, skip_trash: bool = True) -> typing.Sequence[StorageHandler.StorageHandler]:
        storage_handlers = list()
        if directory and directory.exists():
            absolute_file_paths = set()
            for file_path in directory.rglob("*"):
                if not skip_trash or file_path.parent.name != "trash":
                    if not file_path.name.startswith("."):
                        absolute_file_paths.add(str(file_path))
            for file_handler_factory in self._file_handler_factories:
                for data_file in filter(file_handler_factory.is_matching, absolute_file_paths):
                    try:
                        storage_handler = file_handler_factory.make(pathlib.Path(data_file))
                        assert storage_handler.is_valid
                        storage_handlers.append(storage_handler)
                    except Exception as e:
                        logging.error("Exception reading file: %s", data_file)
                        logging.error(str(e))
                        raise
        return storage_handlers


class MemoryStorageHandler(StorageHandler.StorageHandler):

    def __init__(self, uuid_: str, data_properties_map: typing.Dict[str, PersistentDictType],
                 data_map: typing.Dict[str, _NDArray], data_read_event: Event.Event) -> None:
        self.__uuid = uuid_
        self.__data_properties_map = data_properties_map
        self.__data_map = data_map
        self.__data_read_event = data_read_event

    def close(self) -> None:
        self.__uuid = typing.cast(str, None)
        self.__data_properties_map = typing.cast(typing.Dict[str, PersistentDictType], None)
        self.__data_map = typing.cast(typing.Dict[str, _NDArray], None)

    @property
    def factory(self) -> StorageHandler.StorageHandlerFactoryLike:
        return MemoryStorageHandlerFactory()

    @property
    def reference(self) -> str:
        return str(self.__uuid)

    @property
    def is_valid(self) -> bool:
        return True

    def read_properties(self) -> PersistentDictType:
        return copy.deepcopy(self.__data_properties_map.get(self.__uuid, dict()))

    def read_data(self) -> typing.Optional[_NDArray]:
        self.__data_read_event.fire(self.__uuid)
        return self.__data_map.get(self.__uuid)

    def write_properties(self, properties: PersistentDictType, file_datetime: datetime.datetime) -> None:
        self.__data_properties_map[self.__uuid] = Utility.clean_dict(properties)

    def write_data(self, data: _NDArray, file_datetime: datetime.datetime) -> None:
        self.__data_map[self.__uuid] = data.copy()

    def reserve_data(self, data_shape: typing.Tuple[int, ...], data_dtype: numpy.typing.DTypeLike, file_datetime: datetime.datetime) -> None:
        self.__data_map[self.__uuid] = numpy.zeros(data_shape, data_dtype)

    def prepare_move(self) -> None:
        pass

    def remove(self) -> None:
        pass


class MemoryStorageHandlerFactory(StorageHandler.StorageHandlerFactoryLike):

    def is_matching(self, file_path: str) -> bool:
        return True

    def make(self, file_path: pathlib.Path) -> StorageHandler.StorageHandler:
        return MemoryStorageHandler(str(uuid.uuid4()), dict(), dict(), Event.Event())

    def make_path(self, file_path: pathlib.Path) -> str:
        return str(file_path)

    def get_extension(self) -> str:
        return ".memory"


class MemoryProjectStorageSystem(ProjectStorageSystem):

    def __init__(self, *, library_properties: typing.Optional[PersistentDictType] = None,
                 data_properties_map: typing.Optional[typing.Dict[str, PersistentDictType]] = None,
                 data_map: typing.Optional[typing.Dict[str, _NDArray]] = None,
                 trash_map: typing.Optional[typing.Dict[str, PersistentDictType]] = None,
                 data_read_event: typing.Optional[Event.Event] = None) -> None:
        super().__init__()
        self.__library_properties = library_properties if library_properties is not None else dict()
        self.__data_properties_map = data_properties_map if data_properties_map is not None else dict()
        self.__data_map = data_map if data_map is not None else dict()
        self.__trash_map = trash_map if trash_map is not None else dict()
        self._test_data_read_event = data_read_event or Event.Event()
        self._write_count = 0

    def _read_properties(self) -> PersistentDictType:
        return copy.deepcopy(self.__library_properties)

    def _write_properties(self) -> None:
        self._write_count += 1
        self.__library_properties.clear()
        self.__library_properties.update(Model.transform_backward(copy.deepcopy(self.get_storage_properties())))

    def get_identifier(self) -> str:
        return "memory"

    def _make_storage_handler(self, data_item: DataItem.DataItem, file_handler: typing.Optional[StorageHandler.StorageHandlerFactoryLike] = None) -> MemoryStorageHandler:
        data_item_uuid_str = str(data_item.uuid)
        return MemoryStorageHandler(data_item_uuid_str, self.__data_properties_map, self.__data_map, self._test_data_read_event)

    def _find_storage_handlers(self) -> typing.Sequence[StorageHandler.StorageHandler]:
        storage_handlers = list()
        for key in sorted(self.__data_properties_map):
            self.__data_properties_map[key].setdefault("uuid", str(uuid.uuid4()))
            storage_handlers.append(MemoryStorageHandler(key, self.__data_properties_map, self.__data_map, self._test_data_read_event))
        return storage_handlers

    def _is_storage_handler_large_format(self, storage_handler: StorageHandler.StorageHandler) -> bool:
        return False

    def _remove_storage_handler(self, storage_handler: StorageHandler.StorageHandler, *, safe: bool = False) -> None:
        storage_handler_reference = storage_handler.reference
        data = self.__data_map.pop(storage_handler_reference, None)
        properties = self.__data_properties_map.pop(storage_handler_reference)
        if safe:
            assert storage_handler_reference not in self.__trash_map
            self.__trash_map[storage_handler_reference] = {"data": data, "properties": properties}
        storage_handler.close()  # moving files in the storage handler requires it to be closed.

    def _restore_item(self, data_item_uuid: uuid.UUID) -> typing.Optional[PersistentDictType]:
        data_item_uuid_str = str(data_item_uuid)
        trash_entry = self.__trash_map.pop(data_item_uuid_str)
        assert data_item_uuid_str not in self.__data_properties_map
        assert data_item_uuid_str not in self.__data_map
        self.__data_properties_map[data_item_uuid_str] = Migration.transform_to_latest(trash_entry["properties"])
        self.__data_map[data_item_uuid_str] = trash_entry["data"]
        properties = self.__data_properties_map.get(data_item_uuid_str, dict())
        properties["__large_format"] = False
        properties = Migration.transform_to_latest(properties)
        return properties

    def _prune(self) -> None:
        pass  # disabled for testing self.__trash_map = dict()

    def _migrate_data_item(self, reader_info: ReaderInfo, index: int, count: int) -> typing.Optional[ReaderInfo]:
        storage_handler = reader_info.storage_handler
        properties = reader_info.properties
        properties = Utility.clean_dict(copy.deepcopy(properties) if properties else dict())
        if reader_info.changed_ref[0]:
            self.__data_properties_map[storage_handler.reference] = Migration.transform_from_latest(copy.deepcopy(properties))
        return reader_info

    def _migrate_library_properties(self, library_properties: PersistentDictType, reader_info_list: typing.List[ReaderInfo]) -> None:
        self.__library_properties.clear()
        self.__library_properties.update(library_properties)

        data_properties_map = dict()

        for reader_info in reader_info_list:
            data_item_properties = Utility.clean_dict(reader_info.properties if reader_info.properties else dict())
            if data_item_properties.get("version", 0) == DataItem.DataItem.writer_version:
                data_item_properties["__large_format"] = reader_info.large_format
                data_properties_map[reader_info.identifier] = data_item_properties

        def data_item_created(data_item_properties: typing.Tuple[str, PersistentDictType]) -> str:
            # created is a utc timestamp
            earliest_datetime = datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc).isoformat()
            return typing.cast(str, data_item_properties[1].get("created", earliest_datetime))

        data_properties_map = {k: v for k, v in sorted(data_properties_map.items(), key=data_item_created)}

        self.__data_properties_map.clear()
        self.__data_properties_map.update(data_properties_map)

    def _get_migration_stages(self) -> typing.Sequence[ProjectStorageSystemMigrationStage]:
        return [ProjectStorageSystemMigrationStage()]

    def _read_library_properties(self, migration_stage: ProjectStorageSystemMigrationStage) -> PersistentDictType:
        return copy.deepcopy(self.__library_properties)

    def _find_data_items(self, migration_stage: ProjectStorageSystemMigrationStage) -> typing.Sequence[StorageHandler.StorageHandler]:
        return self._find_storage_handlers()


def make_file_persistent_storage_system(path: pathlib.Path) -> Persistence.PersistentStorageInterface:
    return FilePersistentStorageSystem(path)


def make_memory_persistent_storage_system() -> Persistence.PersistentStorageInterface:
    return MemoryPersistentStorageSystem()


def make_index_project_storage_system(project_path: pathlib.Path) -> Persistence.PersistentStorageInterface:
    return FileProjectStorageSystem(project_path)


def make_folder_project_storage_system(project_folder_path: pathlib.Path) -> typing.Optional[Persistence.PersistentStorageInterface]:
    for project_file, project_dir in FileProjectStorageSystem._get_migration_paths(project_folder_path):
        if project_file.exists():
            return FileProjectStorageSystem(project_file, project_dir)
    return None


def make_memory_project_storage_system(profile_context: typing.Any, _uuid: uuid.UUID, d: PersistentDictType) -> typing.Optional[Persistence.PersistentStorageInterface]:
    storage_system_uuid = d.get("uuid") or _uuid or uuid.uuid4()  # allow d to override project uuid for testing failures
    # the profile context must be valid here.
    new_project_properties = profile_context.x_project_properties.setdefault(storage_system_uuid,
                                                                             {"version": d.get("version", PROJECT_VERSION),
                                                                              "uuid": str(storage_system_uuid)})
    new_data_properties_map = profile_context.x_data_properties_map.setdefault(storage_system_uuid, dict())
    new_data_map = profile_context.x_data_map.setdefault(storage_system_uuid, dict())
    new_trash_map = profile_context.x_trash_map.setdefault(storage_system_uuid, dict())
    library_properties = d.get("project_properties", new_project_properties)
    data_properties_map = d.get("data_properties_map", new_data_properties_map)
    data_map = d.get("data_map", new_data_map)
    trash_map = d.get("trash_map", new_trash_map)
    if profile_context.project_properties is None:
        profile_context.project_properties = library_properties
    if profile_context.data_properties_map is None:
        profile_context.data_properties_map = data_properties_map
    if profile_context.data_map is None:
        profile_context.data_map = data_map
    if profile_context.trash_map is None:
        profile_context.trash_map = trash_map
    return MemoryProjectStorageSystem(library_properties=library_properties, data_properties_map=data_properties_map, data_map=data_map, trash_map=trash_map)



"""
Stores application data.
"""

# standard libraries
import copy
import json
import pathlib
import threading
import typing

# third party libraries
from nion.swift.model import Utility
from nion.utils import Event
from nion.utils import StructuredModel


class ApplicationData:
    """Application data is a singleton that stores application data."""

    def __init__(self, file_path: typing.Optional[pathlib.Path] = None) -> None:
        self.__lock = threading.RLock()
        self.__file_path = file_path
        self.__data_dict: typing.Optional[typing.Dict[str, typing.Any]] = None
        self.data_changed_event = Event.Event()

    @property
    def file_path(self) -> typing.Optional[pathlib.Path]:
        return self.__file_path

    @file_path.setter
    def file_path(self, value: pathlib.Path) -> None:
        self.__file_path = value

    def get_data_dict(self) -> typing.Dict[str, typing.Any]:
        with self.__lock:
            data_changed = self.__read_data_dict()
            result = copy.deepcopy(self.__data_dict) if self.__data_dict else dict()
        if data_changed:
            self.data_changed_event.fire()
        return result

    def set_data_dict(self, d: typing.Mapping[str, typing.Any]) -> None:
        with self.__lock:
            self.__data_dict = dict(d)
            self.__write_data_dict()
        self.data_changed_event.fire()

    def __read_data_dict(self) -> bool:
        if self.__data_dict is None and self.__file_path and self.__file_path.exists():
            with open(self.__file_path) as f:
                self.__data_dict = json.load(f)
                return True
        return False

    def __write_data_dict(self) -> None:
        if self.__file_path:
            with Utility.AtomicFileWriter(self.__file_path) as fp:
                json.dump(self.__data_dict, fp, skipkeys=True, indent=4)


__application_data = ApplicationData()


def set_file_path(file_path: pathlib.Path) -> None:
    __application_data.file_path = file_path
    # requires special bootstrapping for session metadata
    __session_metadata.update()


def get_data() -> typing.Dict[str, typing.Any]:
    return __application_data.get_data_dict()


def set_data(d: typing.Mapping[str, typing.Any]) -> None:
    __application_data.set_data_dict(d)


#

class SessionMetadata:
    """Session data is a singleton that stores application data via the ApplicationData singleton."""

    def __init__(self) -> None:
        site_field = StructuredModel.define_field("site", StructuredModel.STRING)
        instrument_field = StructuredModel.define_field("instrument", StructuredModel.STRING)
        task_field = StructuredModel.define_field("task", StructuredModel.STRING)
        microscopist_field = StructuredModel.define_field("microscopist", StructuredModel.STRING)
        sample_field = StructuredModel.define_field("sample", StructuredModel.STRING)
        sample_area_field = StructuredModel.define_field("sample_area", StructuredModel.STRING)

        fields = [site_field, instrument_field, task_field, microscopist_field, sample_field, sample_area_field]
        self.__field_descriptions = typing.cast(typing.List[typing.Mapping[str, typing.Any]], fields)

        schema = StructuredModel.define_record("SessionMetadata", fields)

        self.__model = typing.cast(StructuredModel.RecordModel, StructuredModel.build_model(schema, value=get_data().get("session_metadata", dict())))

        def model_changed() -> None:
            data = get_data()
            data["session_metadata"] = self.__model.to_dict_value()
            set_data(data)

        self.__model_changed_listener = self.__model.model_changed_event.listen(model_changed)

    @property
    def model(self) -> StructuredModel.RecordModel:
        return self.__model

    def update(self) -> None:
        # handle special bootstrapping for session metadata. this can be solved in the long
        # run by using a data structure which properly notifies listeners of changes when reloaded.
        field_names = [fd["name"] for fd in self.__field_descriptions]
        fields = get_data().get("session_metadata", dict())
        for field in fields:
            if field in field_names:
                setattr(self.__model, field, fields[field])


__session_metadata = SessionMetadata()

def get_session_metadata_model() -> StructuredModel.RecordModel:
    return __session_metadata.model

def get_session_metadata_dict() -> typing.Dict[str, typing.Any]:
    return dict(typing.cast(typing.Mapping[str, typing.Any], __session_metadata.model.to_dict_value()))



"""Deprecated. Hardware source has been moved to nionswift-instrumentation-kit.

For backwards compatibility, a few functions are provided until all packages can be migrated to
the new API available in nionswift-instrumentation-kit.
"""
import contextlib
import typing

import numpy.typing

from nion.data import DataAndMetadata
from nion.utils import Registry

_NDArray = numpy.typing.NDArray[typing.Any]
FrameParametersDictType = typing.Dict[str, typing.Any]


class ViewTaskLike(typing.Protocol):
    def close(self) -> None: ...
    def grab_immediate(self) -> typing.Sequence[DataAndMetadata.DataAndMetadata]: ...
    def grab_next_to_finish(self) -> typing.Sequence[DataAndMetadata.DataAndMetadata]: ...
    def grab_next_to_start(self) -> typing.Sequence[DataAndMetadata.DataAndMetadata]: ...
    def grab_earliest(self) -> typing.Sequence[DataAndMetadata.DataAndMetadata]: ...


class FrameParametersLike(typing.Protocol):
    def as_dict(self) -> FrameParametersDictType: ...


class RecordTaskLike(typing.Protocol):

    @property
    def is_started(self) -> bool:
        raise NotImplementedError()

    @property
    def is_finished(self) -> bool:
        raise NotImplementedError()

    def wait_started(self, *, timeout: typing.Optional[float] = None) -> None:
        raise NotImplementedError()

    def wait_finished(self, *, timeout: typing.Optional[float] = None) -> None:
        raise NotImplementedError()

    def grab_xdatas(self, *, timeout: typing.Optional[float] = None) -> typing.Sequence[typing.Optional[DataAndMetadata.DataAndMetadata]]:
        raise NotImplementedError()


class HardwareSourceLike(typing.Protocol):
    @property
    def hardware_source_id(self) -> str: raise NotImplementedError()

    @property
    def selected_profile_index(self) -> int: raise NotImplementedError()

    @property
    def is_playing(self) -> bool: raise NotImplementedError()

    @property
    def is_recording(self) -> bool: raise NotImplementedError()

    def get_record_frame_parameters(self) -> FrameParametersLike: ...
    def set_record_frame_parameters(self, frame_parameters: FrameParametersLike) -> None: ...
    def get_current_frame_parameters(self) -> FrameParametersLike: ...
    def set_current_frame_parameters(self, frame_parameters: FrameParametersLike) -> None: ...
    def get_frame_parameters(self, index: int) -> FrameParametersLike: ...
    def set_frame_parameters(self, index: int, frame_parameters: FrameParametersLike) -> None: ...
    def get_frame_parameters_from_dict(self, d: typing.Mapping[str, typing.Any]) -> FrameParametersLike: ...
    def set_channel_enabled(self, channel_index: int, enabled: bool) -> bool: ...
    def start_playing(self) -> None: ...
    def abort_playing(self) -> None: ...
    def stop_playing(self) -> None: ...
    def start_recording(self) -> RecordTaskLike: ...
    def abort_recording(self) -> None: ...
    def get_next_xdatas_to_finish(self, timeout: typing.Optional[float] = None) -> typing.Sequence[typing.Optional[DataAndMetadata.DataAndMetadata]]: ...
    def get_next_xdatas_to_start(self, timeout: typing.Optional[float] = None) -> typing.Sequence[typing.Optional[DataAndMetadata.DataAndMetadata]]: ...
    def create_view_task(self, frame_parameters: typing.Optional[FrameParametersLike] = None, channels_enabled: typing.Optional[typing.Sequence[bool]] = None, buffer_size: int = 1) -> ViewTaskLike: ...
    def set_selected_profile_index(self, index: int) -> None: ...
    def get_property(self, name: str) -> typing.Any: ...
    def set_property(self, name: str, value: typing.Any) -> None: ...


class InstrumentLike(typing.Protocol):
    @property
    def instrument_id(self) -> str: raise NotImplementedError()

    def set_control_output(self, name: str, value: typing.Any, options: typing.Optional[typing.Mapping[str, typing.Any]] = None) -> None: ...
    def get_control_output(self, name: str) -> float: ...
    def get_control_state(self, name: str) -> str: ...
    def get_property(self, name: str) -> typing.Any: ...
    def set_property(self, name: str, value: typing.Any) -> None: ...


class HardwareSourceManagerInterface(typing.Protocol):
    def register_hardware_source(self, hardware_source: HardwareSourceLike) -> None: ...
    def get_all_instrument_ids(self) -> typing.List[str]: ...
    def get_all_hardware_source_ids(self) -> typing.List[str]: ...
    def get_instrument_by_id(self, instrument_id: str) -> InstrumentLike: ...
    def get_hardware_source_for_hardware_source_id(self, hardware_source_id: str) -> typing.Optional[HardwareSourceLike]: ...
    def make_delegate_hardware_source(self, delegate: typing.Any, hardware_source_id: str, hardware_source_name: str) -> HardwareSourceLike: ...


def hardware_source_manager() -> HardwareSourceManagerInterface:
    return typing.cast(HardwareSourceManagerInterface, Registry.get_component("hardware_source_manager"))


def HardwareSourceManager() -> HardwareSourceManagerInterface:
    return hardware_source_manager()


@contextlib.contextmanager
def get_data_generator_by_id(hardware_source_id: str, sync: bool = True) -> typing.Any:
    """Return a generator for data.

    :param bool sync: whether to wait for current frame to finish then collect next frame

    NOTE: a new ndarray is created for each call.
    """
    hardware_source = HardwareSourceManager().get_hardware_source_for_hardware_source_id(hardware_source_id)
    assert hardware_source

    def get_last_data() -> _NDArray:
        assert hardware_source
        xdata0 = hardware_source.get_next_xdatas_to_finish()[0]
        data = xdata0.data if xdata0 else None
        assert data is not None
        return data.copy()

    yield get_last_data



from __future__ import annotations

# standard libraries
import collections
import gettext
import typing
import uuid

# third party libraries
# None

# local libraries
from nion.swift.model import DisplayItem
from nion.swift.model import Persistence

if typing.TYPE_CHECKING:
    from nion.swift.model import Project

_SpecifierType = Persistence._SpecifierType
_SpecifierDictType = Persistence._SpecifierDictType


_ = gettext.gettext


"""

    Group layout.
    Live. Live acquisition items.
    Library. Static items.

    *   The user can edit/delete the Library group, and create additional top level groups.
    However, at least one group in addition to the Live group must always be present.
    *   The user cannot delete the live group.
    *   The user cannot add/delete items from the Live group.
    *   Groups can be re-ordered, but the Live group is always first.
    *   New items are put in the first group after Live.

    *   NOTES
    Groups can contain child groups and data items.

    One design option is to have the document controller hold a single data group.
    This has the advantage of making access very uniform. However, this also makes
    it very easy for the user to "lose" items at the top level, which they might not
    see unless they selected the topmost group.

    Another design option is to have the document controller hold a list of data
    groups. This is more complex structure, but makes it easier for the user to
    understand the structure.

    Smart Groups

    Smart groups work on sibling data groups (aka the data items in the smart group's
    container). A smart group at the top level of groups will be able to filter
    and sort data items contained in other top level groups.
    Is there a need for smart groups anywhere except at the top level?

    - (A1a: 1, A1b: 1, A1b1: 2, A1b1a: 2, A1b2: 1, B1: 1, B1a: 1) [0]
    Smart Group [-]
    Group A (A1a: 1, A1b: 1, A1b1: 1, A1b1a: 1, A1b2: 1) [-]
    Group A1 (A1a: 1, A1b: 1, A1b1: 1, A1b1a: 1, A1b2: 1) [A]
    Data Item A1a () [A1]
    Data Item A1b (A1b1: 1, A1b1a: 1, A1b2: 1) [A1]
    Data Item A1b1 (A1b1a: 1 [A1b]
    Data Item A1b1a () [A1b1]
    Data Item A1b2 () [A1b]
    Group B (B1: 1, B1a: 1, A1b1: 1, A1b1a: 1) [-]
    Data Item B1 (B1a: 1) [B]
    Data Item B1a [B1]
    Data Item A1b1 (A1b1a: 1) [B, A1b]
    Data Item A1b1a [A1b1]

    Data item A1b1b gets added to A1b1. A1b1 counted_display_items gets updated, then tells its containers B, A1b
    that they have been updated. B, A1b counted_display_items get updated, then tell their containers, etc.
    When each group gets the counted_display_items_updated message, it tells any smart container children
    that they need to re-filter.

    Data item A2 gets added to Group A.

    """


class UuidsToStringsConverter:
    def convert(self, value: typing.List[uuid.UUID]) -> typing.List[str]:
        return [str(uuid_) for uuid_ in value]

    def convert_back(self, value: typing.List[str]) -> typing.List[uuid.UUID]:
        return [uuid.UUID(uuid_str) for uuid_str in value]


class DataGroup(Persistence.PersistentObject):

    def __init__(self) -> None:
        super().__init__()
        self.define_type("data_group")
        self.define_property("title", _("Untitled"), hidden=True, validate=self.__validate_title, changed=self.__property_changed)
        self.define_property("display_item_specifiers", list(), hidden=True, validate=self.__validate_display_item_specifiers, changed=self.__property_changed, key="display_item_references")
        self.define_relationship("data_groups", data_group_factory, insert=self.__insert_data_group, remove=self.__remove_data_group, hidden=True)
        self.__lookup_display_item: typing.Optional[typing.Callable[[_SpecifierType], typing.Optional[DisplayItem.DisplayItem]]] = None
        self.__display_items: typing.List[DisplayItem.DisplayItem] = list()
        # Python 3.9+: typed counter
        self.__counted_display_items = collections.Counter()  # type: ignore

    def __str__(self) -> str:
        return self.title

    @property
    def title(self) -> str:
        return typing.cast(str, self._get_persistent_property_value("title"))

    @title.setter
    def title(self, value: str) -> None:
        self._set_persistent_property_value("title", value)

    @property
    def display_item_specifiers(self) -> typing.List[_SpecifierType]:
        return typing.cast(typing.List[_SpecifierType], self._get_persistent_property_value("display_item_specifiers"))

    @display_item_specifiers.setter
    def display_item_specifiers(self, value: typing.Sequence[_SpecifierType]) -> None:
        self._set_persistent_property_value("display_item_specifiers", list(value))

    @property
    def data_groups(self) -> typing.Sequence[DataGroup]:
        return typing.cast(typing.Sequence[DataGroup], self._get_relationship_values("data_groups"))

    @property
    def project(self) -> Project.Project:
        if isinstance(self.container, self.__class__):
            return self.container.project
        return typing.cast("Project.Project", self.container)

    def create_proxy(self) -> Persistence.PersistentObjectProxy[DataGroup]:
        return self.project.create_item_proxy(item=self)

    @property
    def item_specifier(self) -> Persistence.PersistentObjectSpecifier:
        return Persistence.PersistentObjectSpecifier(self.uuid)

    def __validate_title(self, value: typing.Any) -> str:
        return str(value) if value is not None else str()

    def __validate_display_item_specifiers(self, display_item_specifiers: typing.Sequence[_SpecifierType]) -> typing.Sequence[_SpecifierType]:
        # remove duplicates
        new_display_item_specifiers = list()
        for display_item_specifier in display_item_specifiers:
            if display_item_specifier not in new_display_item_specifiers:
                new_display_item_specifiers.append(display_item_specifier)
        return new_display_item_specifiers

    def __property_changed(self, name: str, value: typing.Any) -> None:
        self.notify_property_changed(name)

    def connect_display_items(self, lookup_display_item: typing.Callable[[_SpecifierType], typing.Optional[DisplayItem.DisplayItem]]) -> None:
        for data_group in self.data_groups:
            data_group.connect_display_items(lookup_display_item)
        display_items: typing.List[DisplayItem.DisplayItem] = list()
        for display_item_specifier in self.display_item_specifiers:
            display_item = lookup_display_item(display_item_specifier)
            if display_item:
                display_items.append(display_item)
        self.__display_items = display_items
        self.__lookup_display_item = lookup_display_item

    def disconnect_display_items(self) -> None:
        for data_group in self.data_groups:
            data_group.disconnect_display_items()
        self.__lookup_display_item = None

    def append_display_item(self, display_item: DisplayItem.DisplayItem) -> None:
        self.insert_display_item(len(self.__display_items), display_item)

    def insert_display_item(self, before_index: int, display_item: DisplayItem.DisplayItem) -> None:
        assert display_item not in self.__display_items
        self.__display_items.insert(before_index, display_item)
        self.notify_insert_item("display_items", display_item, before_index)
        self.update_counted_display_items(collections.Counter([display_item]))
        display_item_specifiers = self.display_item_specifiers
        display_item_specifier = Persistence.write_persistent_specifier(display_item.uuid)
        assert display_item_specifier is not None
        display_item_specifiers.insert(before_index, display_item_specifier)
        self.display_item_specifiers = display_item_specifiers
        self.notify_property_changed("display_item_specifiers")

    def remove_display_item(self, display_item: DisplayItem.DisplayItem) -> None:
        index = self.__display_items.index(display_item)
        self.__display_items.remove(display_item)
        self.subtract_counted_display_items(collections.Counter([display_item]))
        self.notify_remove_item("display_items", display_item, index)
        display_item_specifiers = self.display_item_specifiers
        display_item_specifiers.pop(index)
        self.display_item_specifiers = display_item_specifiers
        self.notify_property_changed("display_item_specifiers")

    @property
    def display_items(self) -> typing.Tuple[DisplayItem.DisplayItem, ...]:
        return tuple(self.__display_items)

    def append_data_group(self, data_group: DataGroup) -> None:
        self.insert_data_group(len(self.data_groups), data_group)

    def insert_data_group(self, before_index: int, data_group: DataGroup) -> None:
        self.insert_item("data_groups", before_index, data_group)

    def remove_data_group(self, data_group: DataGroup) -> None:
        self.remove_item("data_groups", data_group)

    def __insert_data_group(self, name: str, before_index: int, data_group: DataGroup) -> None:
        if self.__lookup_display_item:
            data_group.connect_display_items(self.__lookup_display_item)
        self.update_counted_display_items(data_group.counted_display_items)
        self.notify_insert_item("data_groups", data_group, before_index)

    def __remove_data_group(self, name: str, index: int, data_group: DataGroup) -> None:
        data_group.disconnect_display_items()
        self.subtract_counted_display_items(data_group.counted_display_items)
        self.notify_remove_item("data_groups", data_group, index)

    @property
    def counted_display_items(self) -> collections.Counter[DisplayItem.DisplayItem]:
        return self.__counted_display_items

    def update_counted_display_items(self, counted_display_items: collections.Counter[DisplayItem.DisplayItem]) -> None:
        self.__counted_display_items.update(counted_display_items)

    def subtract_counted_display_items(self, counted_display_items: collections.Counter[DisplayItem.DisplayItem]) -> None:
        self.__counted_display_items.subtract(counted_display_items)
        self.__counted_display_items += collections.Counter()  # strip empty items


class _DataGroupContainer(typing.Protocol):

    @property
    def data_groups(self) -> typing.Sequence[DataGroup]: raise NotImplementedError()

    @property
    def display_items(self) -> typing.Sequence[DisplayItem.DisplayItem]: raise NotImplementedError()


# return a generator for all data groups and child data groups in container
def get_flat_data_group_generator_in_container(container: _DataGroupContainer) -> typing.Iterator[DataGroup]:
    for data_group in container.data_groups:
        yield data_group
        for child_data_group in get_flat_data_group_generator_in_container(data_group):
            yield child_data_group


# return a generator for all data items, child data items, and data items in child groups in container
def get_flat_display_item_generator_in_container(container: _DataGroupContainer) -> typing.Iterator[DisplayItem.DisplayItem]:
    if hasattr(container, "display_items"):
        for display_item in container.display_items:
            yield display_item
    if hasattr(container, "data_groups"):
        for data_group in container.data_groups:
            for display_item in get_flat_display_item_generator_in_container(data_group):
                yield display_item


# Return the data_group matching name that is the descendant of the container.
def get_data_group_in_container_by_title(container: _DataGroupContainer, data_group_title: str) -> typing.Optional[DataGroup]:
    for data_group in container.data_groups:
        if data_group.title == data_group_title:
            return data_group
    return None


def data_group_factory(lookup_id: typing.Callable[[str], str]) -> DataGroup:
    return DataGroup()



from __future__ import annotations

# standard libraries
import asyncio
import dataclasses
import functools
import gettext
import operator
import threading
import typing
import weakref

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import Core
from nion.data import DataAndMetadata
from nion.data import Image
from nion.swift import Panel
from nion.swift.model import DisplayItem
from nion.swift.model import Graphics
from nion.ui import CanvasItem
from nion.ui import DrawingContext
from nion.ui import Widgets
from nion.utils import Binding
from nion.utils import Event
from nion.utils import Model
from nion.utils import Observable
from nion.utils import Process
from nion.utils import ReferenceCounting
from nion.utils import Stream

if typing.TYPE_CHECKING:
    from nion.swift import DocumentController
    from nion.swift.model import Persistence
    from nion.ui import UserInterface

_RGBA8ImageDataType = Image._RGBA8ImageDataType

_NDArray = numpy.typing.NDArray[typing.Any]

_StatisticsTable = typing.Dict[str, str]

_ = gettext.gettext

T = typing.TypeVar('T')
IT = typing.TypeVar('IT')
OT = typing.TypeVar('OT')


class AdornmentsCanvasItem(CanvasItem.AbstractCanvasItem):
    """A canvas item to draw the adornments on top of the histogram.

    The adornments are the black and white lines shown during mouse
     adjustment of the display limits.

    Callers are expected to set the display_limits property and
     then call update.
    """

    def __init__(self) -> None:
        super().__init__()
        self.display_limits: typing.Tuple[float, float] = (0.0, 1.0)

    def _repaint(self, drawing_context: DrawingContext.DrawingContext) -> None:
        """Repaint the canvas item. This will occur on a thread."""

        # canvas size
        canvas_size = self.canvas_size
        if canvas_size:
            left = self.display_limits[0]
            right = self.display_limits[1]

            # draw left display limit
            if left > 0.0:
                with drawing_context.saver():
                    drawing_context.begin_path()
                    drawing_context.move_to(left * canvas_size.width, 1)
                    drawing_context.line_to(left * canvas_size.width, canvas_size.height - 1)
                    drawing_context.line_width = 2
                    drawing_context.stroke_style = "#000"
                    drawing_context.stroke()

            # draw right display limit
            if right < 1.0:
                with drawing_context.saver():
                    drawing_context.begin_path()
                    drawing_context.move_to(right * canvas_size.width, 1)
                    drawing_context.line_to(right * canvas_size.width, canvas_size.height - 1)
                    drawing_context.line_width = 2
                    drawing_context.stroke_style = "#FFF"
                    drawing_context.stroke()

            # draw border
            with drawing_context.saver():
                drawing_context.begin_path()
                drawing_context.move_to(0, canvas_size.height)
                drawing_context.line_to(canvas_size.width, canvas_size.height)
                drawing_context.line_width = 1
                drawing_context.stroke_style = "#444"
                drawing_context.stroke()


class SimpleLineGraphCanvasItem(CanvasItem.AbstractCanvasItem):
    """A canvas item to draw a simple line graph.

    The caller can specify a background color by setting the background_color
     property in the format of a CSS color.

    The caller must update the data by setting the data property. The data must
     be a numpy array with a range from 0,1. The data will be re-binned to the
     width of the canvas item and plotted.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__data: typing.Optional[_NDArray] = None
        self.__background_color: typing.Optional[typing.Union[str, DrawingContext.LinearGradient]] = None
        self.__retained_rebin_1d: typing.Dict[str, typing.Any] = dict()

    @property
    def data(self) -> typing.Optional[_NDArray]:
        """Return the data."""
        return self.__data

    @data.setter
    def data(self, data: typing.Optional[_NDArray]) -> None:
        """Set the data and mark the canvas item for updating.

        Data should be a numpy array with a range from 0,1.
        """
        self.__data = data
        self.update()

    @property
    def background_color(self) -> typing.Optional[typing.Union[str, DrawingContext.LinearGradient]]:
        """Return the background color."""
        return self.__background_color

    @background_color.setter
    def background_color(self, background_color: typing.Optional[typing.Union[str, DrawingContext.LinearGradient]]) -> None:
        """Set the background color. Use CSS color format."""
        self.__background_color = background_color
        self.update()

    def _repaint(self, drawing_context: DrawingContext.DrawingContext) -> None:
        """Repaint the canvas item. This will occur on a thread."""
        canvas_size = self.canvas_size
        if canvas_size:
            # draw background
            if self.background_color:
                with drawing_context.saver():
                    drawing_context.begin_path()
                    drawing_context.move_to(0, 0)
                    drawing_context.line_to(canvas_size.width, 0)
                    drawing_context.line_to(canvas_size.width, canvas_size.height)
                    drawing_context.line_to(0, canvas_size.height)
                    drawing_context.close_path()
                    drawing_context.fill_style = self.background_color
                    drawing_context.fill()

            # draw the data, if any
            if (self.data is not None and len(self.data) > 0):

                # draw the histogram itself
                with drawing_context.saver():
                    drawing_context.begin_path()
                    binned_data = Image.rebin_1d(self.data, int(canvas_size.width), self.__retained_rebin_1d) if int(canvas_size.width) != self.data.shape[0] else self.data
                    for i in range(canvas_size.width):
                        drawing_context.move_to(i, canvas_size.height)
                        drawing_context.line_to(i, canvas_size.height * (1 - binned_data[i]))
                    drawing_context.line_width = 1
                    drawing_context.stroke_style = "#444"
                    drawing_context.stroke()


class ColorMapCanvasItem(CanvasItem.AbstractCanvasItem):

    def __init__(self) -> None:
        super().__init__()
        self.update_sizing(self.sizing.with_fixed_height(4))
        self.__color_map_data: typing.Optional[_RGBA8ImageDataType] = None

    @property
    def color_map_data(self) -> typing.Optional[_RGBA8ImageDataType]:
        """Return the data."""
        return self.__color_map_data

    @color_map_data.setter
    def color_map_data(self, data: typing.Optional[_RGBA8ImageDataType]) -> None:
        """Set the data and mark the canvas item for updating.

        Data should be an ndarray of shape (256, 3) with type uint8
        """
        self.__color_map_data = data
        self.update()

    def _repaint(self, drawing_context: DrawingContext.DrawingContext) -> None:
        """Repaint the canvas item. This will occur on a thread."""
        canvas_size = self.canvas_size
        if canvas_size:
            with drawing_context.saver():
                if self.__color_map_data is not None:
                    rgba_image: numpy.typing.NDArray[numpy.uint32] = numpy.empty((4,) + self.__color_map_data.shape[:-1], dtype=numpy.uint32)
                    Image.get_rgb_view(rgba_image)[:] = self.__color_map_data[numpy.newaxis, :, :]  # scalar data assigned to each component of rgb view
                    Image.get_alpha_view(rgba_image)[:] = 255
                    drawing_context.draw_image(rgba_image, 0, 0, canvas_size.width, canvas_size.height)


class HistogramCanvasItem(CanvasItem.CanvasItemComposition):
    """A canvas item to draw and control a histogram."""

    def __init__(self, cursor_changed_fn: typing.Callable[[typing.Optional[float]], None]) -> None:
        super().__init__()

        # tell the canvas item that we want mouse events.
        self.wants_mouse_events = True

        # create the component canvas items: adornments and the graph.
        self.__adornments_canvas_item = AdornmentsCanvasItem()
        self.__simple_line_graph_canvas_item = SimpleLineGraphCanvasItem()
        self.__histogram_color_map_canvas_item = ColorMapCanvasItem()

        # canvas items get added back to front

        column = CanvasItem.CanvasItemComposition()
        column.layout = CanvasItem.CanvasItemColumnLayout()

        graph_and_adornments = CanvasItem.CanvasItemComposition()
        graph_and_adornments.add_canvas_item(self.__simple_line_graph_canvas_item)
        graph_and_adornments.add_canvas_item(self.__adornments_canvas_item)

        column.add_canvas_item(graph_and_adornments)
        column.add_canvas_item(self.__histogram_color_map_canvas_item)

        self.add_canvas_item(column)

        # used for mouse tracking.
        self.__pressed = False

        self.on_set_display_limits: typing.Optional[typing.Callable[[typing.Optional[typing.Tuple[float, float]]], None]] = None

        self.__cursor_changed = cursor_changed_fn

    def close(self) -> None:
        self._set_histogram_data(None)
        super().close()

    @property
    def background_color(self) -> typing.Optional[typing.Union[str, DrawingContext.LinearGradient]]:
        """Return the background color."""
        return self.__simple_line_graph_canvas_item.background_color

    @background_color.setter
    def background_color(self, background_color: typing.Optional[typing.Union[str, DrawingContext.LinearGradient]]) -> None:
        """Set the background color, in the CSS color format."""
        self.__simple_line_graph_canvas_item.background_color = background_color

    def _set_histogram_data(self, histogram_data: typing.Optional[_NDArray]) -> None:
        # if the user is currently dragging the display limits, we don't want to update
        # from changing data at the same time. but we _do_ want to draw the updated data.
        if not self.__pressed:
            self.__adornments_canvas_item.display_limits = (0, 1)

        self.histogram_data = histogram_data

        # make sure the adornments get updated
        self.__adornments_canvas_item.update()

    @property
    def histogram_data(self) -> typing.Optional[_NDArray]:
        return self.__simple_line_graph_canvas_item.data

    @histogram_data.setter
    def histogram_data(self, histogram_data: typing.Optional[_NDArray]) -> None:
        self.__simple_line_graph_canvas_item.data = histogram_data

    @property
    def color_map_data(self) -> typing.Optional[_RGBA8ImageDataType]:
        return self.__histogram_color_map_canvas_item.color_map_data

    @color_map_data.setter
    def color_map_data(self, color_map_data: typing.Optional[_RGBA8ImageDataType]) -> None:
        self.__histogram_color_map_canvas_item.color_map_data = color_map_data

    def __set_display_limits(self, display_limits: typing.Tuple[float, float]) -> None:
        self.__adornments_canvas_item.display_limits = display_limits
        self.__adornments_canvas_item.update()

    def mouse_double_clicked(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        if super().mouse_double_clicked(x, y, modifiers):
            return True
        self.__set_display_limits((0, 1))
        if callable(self.on_set_display_limits):
            self.on_set_display_limits(None)
        return True

    def mouse_pressed(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        if super().mouse_pressed(x, y, modifiers):
            return True
        canvas_size = self.canvas_size
        if canvas_size:
            self.__pressed = True
            self.start = float(x) / canvas_size.width
            self.__set_display_limits((self.start, self.start))
            return True
        return False

    def mouse_released(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        if super().mouse_released(x, y, modifiers):
            return True
        self.__pressed = False
        display_limit_range = self.__adornments_canvas_item.display_limits[1] - self.__adornments_canvas_item.display_limits[0]
        if 0 < display_limit_range < 1:
            if callable(self.on_set_display_limits):
                self.on_set_display_limits(self.__adornments_canvas_item.display_limits)
        self.__set_display_limits((0, 1))
        return True

    def mouse_position_changed(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        canvas_size = self.canvas_size
        if canvas_size:
            if callable(self.__cursor_changed):
                self.__cursor_changed(x / canvas_size.width)
            if super().mouse_position_changed(x, y, modifiers):
                return True
            if self.__pressed:
                current = float(x) / canvas_size.width
                self.__set_display_limits((min(self.start, current), max(self.start, current)))
            return True
        return False

    def mouse_exited(self) -> bool:
        if callable(self.__cursor_changed):
            self.__cursor_changed(None)
        return True


@dataclasses.dataclass
class HistogramWidgetData:
    data: typing.Optional[_NDArray] = None
    display_range: typing.Optional[typing.Tuple[float, float]] = None

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, self.__class__):
            return numpy.array_equal(self.data, other.data) and self.display_range == other.display_range  # type: ignore
        return False


class HistogramWidget(Widgets.CompositeWidgetBase):

    def __init__(self, document_controller: DocumentController.DocumentController,
                 display_item_stream: Stream.AbstractStream[DisplayItem.DisplayItem],
                 histogram_widget_data_model: Model.PropertyModel[HistogramWidgetData],
                 color_map_data_model: Model.PropertyModel[_RGBA8ImageDataType],
                 cursor_changed_fn: typing.Callable[[typing.Optional[float], typing.Optional[typing.Tuple[float, float]]], None]) -> None:
        content_widget = document_controller.ui.create_column_widget(properties={"min-height": 84, "max-height": 84})
        super().__init__(content_widget)

        ui = document_controller.ui

        self.__display_item_stream = display_item_stream.add_ref()

        self.__histogram_data_model = histogram_widget_data_model
        self.__color_map_data_model = color_map_data_model

        self.__display_range: typing.Optional[typing.Tuple[float, float]] = None

        def histogram_data_changed(key: str) -> None:
            if key == "value":
                histogram_widget_data = self.__histogram_data_model.value
                if histogram_widget_data:
                    self.__histogram_canvas_item._set_histogram_data(histogram_widget_data.data)
                    self.__display_range = histogram_widget_data.display_range

        self.__histogram_data_property_changed_event_listener = self.__histogram_data_model.property_changed_event.listen(histogram_data_changed)

        def set_display_limits(display_limits: typing.Optional[typing.Tuple[float, float]]) -> None:
            # display_limits in this context are in the range of 0,1
            # we ask for the display_range from the display to get actual
            # data values (never None), and create new display limits
            # based on those data values combined with display_limits.
            # then we set the display_limits on the display, which have
            # the same units as the data values.
            display_item = self.__display_item_stream.value
            display_data_channel = display_item.display_data_channel if display_item else None
            if display_data_channel:
                new_display_limits = None
                if display_limits is not None and self.__display_range is not None:
                    data_min, data_max = self.__display_range
                    lower_display_limit = data_min + display_limits[0] * (data_max - data_min)
                    upper_display_limit = data_min + display_limits[1] * (data_max - data_min)
                    new_display_limits = (lower_display_limit, upper_display_limit)

                action_context = document_controller._get_action_context()
                action_context.parameters["display_limits"] = new_display_limits
                document_controller.perform_action_in_context("raster_display.set_display_limits", action_context)

        def cursor_changed(canvas_x: typing.Optional[float]) -> None:
            if callable(cursor_changed_fn):
                cursor_changed_fn(canvas_x, self.__display_range)

        # create a canvas widget for this panel and put a histogram canvas item in it.
        self.__histogram_canvas_item = HistogramCanvasItem(cursor_changed)
        self.__histogram_canvas_item.on_set_display_limits = set_display_limits

        histogram_widget = ui.create_canvas_widget()
        histogram_widget.canvas_item.add_canvas_item(self.__histogram_canvas_item)

        def color_map_data_changed(key: str) -> None:
            if key == "value":
                self.__histogram_canvas_item.color_map_data = self.__color_map_data_model.value

        self.__color_map_data_stream_listener = self.__color_map_data_model.property_changed_event.listen(color_map_data_changed)

        histogram_data_changed("value")

        color_map_data_changed("value")

        content_widget.add(histogram_widget)

    def close(self) -> None:
        self.__color_map_data_stream_listener.close()
        self.__color_map_data_stream_listener = typing.cast(typing.Any, None)
        self.__display_item_stream.remove_ref()
        self.__display_item_stream = typing.cast(typing.Any, None)
        self.__histogram_canvas_item = typing.cast(typing.Any, None)
        self.__histogram_data_property_changed_event_listener.close()
        self.__histogram_data_property_changed_event_listener = typing.cast(typing.Any, None)
        self.__histogram_data_model = typing.cast(typing.Any, None)
        self.__color_map_data_model = typing.cast(typing.Any, None)
        super().close()

    @property
    def _histogram_canvas_item(self) -> HistogramCanvasItem:
        return self.__histogram_canvas_item

    @property
    def _histogram_data_func_value_model(self) -> Model.PropertyModel[HistogramWidgetData]:
        # for testing
        return self.__histogram_data_model


class StatisticsWidget(Widgets.CompositeWidgetBase):

    def __init__(self, ui: UserInterface.UserInterface, statistics_model: Model.PropertyModel[typing.Dict[str, str]]) -> None:
        content_widget = ui.create_column_widget(properties={"min-height": 18 * 3, "max-height": 18 * 3})
        super().__init__(content_widget)

        # create property models for the UI
        self._stats1_property = Model.PropertyModel[str](str())
        self._stats2_property = Model.PropertyModel[str](str())

        self.__statistics_model = statistics_model

        def statistics_changed(key: str) -> None:
            if key == "value":
                statistics_data = self.__statistics_model.value or dict()
                statistic_strings = list()
                for key in sorted(statistics_data.keys()):
                    value = statistics_data[key]
                    if value is not None:
                        statistic_str = "{0} {1}".format(key, value)
                    else:
                        statistic_str = "{0} {1}".format(key, _("N/A"))
                    statistic_strings.append(statistic_str)
                self._stats1_property.value = "\n".join(statistic_strings[:(len(statistic_strings) + 1) // 2])
                self._stats2_property.value = "\n".join(statistic_strings[(len(statistic_strings) + 1) // 2:])

        self.__statistics_property_changed_event_listener = self.__statistics_model.property_changed_event.listen(statistics_changed)

        statistics_changed("value")

        stats_column1 = ui.create_column_widget(properties={"min-width": 140, "max-width": 140})
        stats_column2 = ui.create_column_widget(properties={"min-width": 140, "max-width": 140})
        stats_column1_label = ui.create_label_widget()
        stats_column2_label = ui.create_label_widget()
        stats_column1.add(stats_column1_label)
        stats_column2.add(stats_column2_label)
        stats_section = ui.create_row_widget()
        stats_section.add_spacing(13)
        stats_section.add(stats_column1)
        stats_section.add_stretch()
        stats_section.add(stats_column2)
        stats_section.add_spacing(13)

        stats_column1_label.bind_text(Binding.PropertyBinding(self._stats1_property, "value"))
        stats_column2_label.bind_text(Binding.PropertyBinding(self._stats2_property, "value"))

        content_widget.add(stats_section)

    def close(self) -> None:
        self.__statistics_property_changed_event_listener.close()
        self.__statistics_property_changed_event_listener = typing.cast(typing.Any, None)
        self.__statistics_model = typing.cast(typing.Any, None)
        super().close()


# Python 3.9+: weakref typing
def calculate_region_data(display_data_and_metadata_ref: typing.Any, region_ref: typing.Any) -> typing.Optional[DataAndMetadata.DataAndMetadata]:
    display_data_and_metadata = typing.cast(typing.Optional[DataAndMetadata.DataAndMetadata], display_data_and_metadata_ref() if display_data_and_metadata_ref else None)
    region = typing.cast(typing.Optional[Graphics.Graphic], region_ref() if region_ref else None)
    if region and display_data_and_metadata:
        if display_data_and_metadata.is_data_1d and isinstance(region, Graphics.IntervalGraphic):
            interval = region.interval
            if 0 <= interval[0] < 1 and 0 < interval[1] <= 1:
                start, end = int(interval[0] * display_data_and_metadata.data_shape[0]), int(interval[1] * display_data_and_metadata.data_shape[0])
                if end - start >= 1:
                    cropped_data_and_metadata = Core.function_crop_interval(display_data_and_metadata, interval)
                    if cropped_data_and_metadata:
                        return cropped_data_and_metadata
        elif display_data_and_metadata.is_data_2d and isinstance(region, Graphics.RectangleTypeGraphic):
            cropped_data_and_metadata = Core.function_crop(display_data_and_metadata, region.bounds.as_tuple())
            if cropped_data_and_metadata:
                return cropped_data_and_metadata
    return display_data_and_metadata


def calculate_histogram_widget_data(display_data_and_metadata: typing.Optional[DataAndMetadata.DataAndMetadata], display_range: typing.Optional[typing.Tuple[float, float]]) -> HistogramWidgetData:
    bins = 320
    subsample = 0  # hard coded subsample size
    subsample_fraction = None  # fraction of total pixels
    subsample_min = 1024  # minimum subsample size
    display_data = display_data_and_metadata.data if display_data_and_metadata else None
    display_data_and_metadata = None  # release ref for gc. needed for tests, because this may occur on a thread.
    if display_data is not None:
        total_pixels = numpy.prod(display_data.shape, dtype=numpy.uint64)
        if not subsample and subsample_fraction:
            subsample = min(max(total_pixels * subsample_fraction, subsample_min), total_pixels)
        if subsample:
            data_sample = numpy.random.choice(display_data.reshape(numpy.prod(display_data.shape, dtype=numpy.uint64)), subsample)
        else:
            data_sample = numpy.copy(display_data)
        if display_range is None or data_sample is None:
            return HistogramWidgetData()
        # numpy is slow because it throws out data less/greater than the min/max values
        # the alternate algorithm here takes a different, faster approach and allows the binning
        # to occur; but throws out the data in the first and last bin. this is not as accurate
        # but improves the speed (compared to numpy) by a factor of 10x.
        range_ = (display_range[1] - display_range[0])
        if range_ > 0.0:
            # int clipping seems faster
            histogram_data = numpy.bincount(numpy.clip(((bins + 2) * ((data_sample.ravel() - display_range[0]) / range_)).astype(int), 0, bins + 2), minlength=bins + 2)[1:bins + 1]
            # histogram_data = numpy.bincount(((bins + 2) * numpy.clip((data_sample - display_range[0]) / (display_range[1] - display_range[0]), 0.0, 1.0)).astype(int).ravel())[1:bins+1]
        else:
            histogram_data = numpy.zeros((bins,), dtype=int)
        # why can't numpy make this optimization?
        # histogram_data = factor * numpy.histogram(data_sample, range=display_range, bins=bins)[0]  # type: ignore
        histogram_max = numpy.max(histogram_data)  # assumes that histogram_data is int
        if histogram_max > 0:
            histogram_data = histogram_data / float(histogram_max)  # type: ignore
        return HistogramWidgetData(histogram_data, display_range)
    return HistogramWidgetData()


def calculate_statistics(display_data_and_metadata: typing.Optional[DataAndMetadata.DataAndMetadata], display_data_range: typing.Optional[typing.Tuple[float, float]], region: typing.Optional[Graphics.Graphic], displayed_intensity_calibration: typing.Optional[Calibration.Calibration]) -> _StatisticsTable:
    data = display_data_and_metadata.data if display_data_and_metadata else None
    display_data_and_metadata = None  # release ref for gc. needed for tests, because this may occur on a thread.
    data_range = display_data_range
    if data is not None and data.size > 0 and displayed_intensity_calibration:
        mean = numpy.mean(data).item()
        std = numpy.std(data).item()
        rms = numpy.sqrt(numpy.mean(numpy.square(numpy.absolute(data)))).item()
        dimensional_shape = Image.dimensional_shape_from_shape_and_dtype(data.shape, data.dtype) or (1, 1)
        sum_data = mean * functools.reduce(operator.mul, dimensional_shape)
        if region is None:
            data_min, data_max = data_range if data_range is not None else (None, None)
        else:
            data_min, data_max = numpy.amin(data), numpy.amax(data)
        mean_str = displayed_intensity_calibration.convert_to_calibrated_value_str(mean)
        std_str = displayed_intensity_calibration.convert_to_calibrated_value_str(std)
        data_min_str = displayed_intensity_calibration.convert_to_calibrated_value_str(data_min) if data_min is not None else str()
        data_max_str = displayed_intensity_calibration.convert_to_calibrated_value_str(data_max) if data_max is not None else str()
        rms_str = displayed_intensity_calibration.convert_to_calibrated_value_str(rms)
        sum_data_str = displayed_intensity_calibration.convert_to_calibrated_value_str(sum_data)

        return { "mean": mean_str, "std": std_str, "min": data_min_str, "max": data_max_str, "rms": rms_str, "sum": sum_data_str }
    return dict()


class PropertySetter(typing.Generic[T]):
    def __init__(self, stream: Stream.AbstractStream[T], target: typing.Any, property: str) -> None:
        self.__stream = stream

        # define a stub and use weak_partial to avoid holding references to self.
        def value_changed(value: typing.Optional[T]) -> None:
            setattr(target, property, value)

        self.__stream_listener = self.__stream.value_stream.listen(value_changed)


class HistogramProcessor(Observable.Observable):
    """Computes a histogram and statistics."""

    def __init__(self, event_loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> None:
        super().__init__()
        event_loop = event_loop or asyncio.get_running_loop()
        assert event_loop
        self.__lock = threading.RLock()
        # these fields are used for inputs.
        self.__display_data_and_metadata: typing.Optional[DataAndMetadata.DataAndMetadata] = None
        self.__region: typing.Optional[Graphics.Graphic] = None
        self.__display_range: typing.Optional[typing.Tuple[float, float]] = None
        self.__display_data_range: typing.Optional[typing.Tuple[float, float]] = None
        self.__displayed_intensity_calibration: typing.Optional[Calibration.Calibration] = None
        # these fields are used for computation.
        self.__histogram_widget_data_dirty = False
        self.__statistics_dirty = False
        self.__region_data_and_metadata: typing.Optional[DataAndMetadata.DataAndMetadata] = None
        # these fields are used for outputs.
        self.__histogram_widget_data = HistogramWidgetData()
        self.__statistics: _StatisticsTable = dict()

        self.__event_loop = event_loop
        self.__cancel = threading.Event()
        self.__event = threading.Event()
        self.__handle_lock = threading.RLock()
        self.__handle: typing.Optional[asyncio.Handle] = None
        self.__thread = threading.Thread(target=self.__run)
        self.__thread.start()

    def close(self) -> None:
        with self.__handle_lock:
            if self.__handle:
                self.__handle.cancel()
        self.__cancel.set()
        self.__event.set()
        self.__thread.join(1.0)
        self._thread = None

    def __run(self) -> None:
        while not self.__cancel.is_set():
            if self.__event.wait(0.05):  # frequency to check for cancel
                self.__event.clear()
                if self.__cancel.wait(0.25):  # frequency to wait between updates
                    break
                with Process.audit("histogram"):
                    old_histogram_widget_data = self.__histogram_widget_data
                    old_statistics = self.__statistics
                    self.__evaluate()
                    notify_data = old_histogram_widget_data != self.__histogram_widget_data
                    notify_statistics = old_statistics != self.__statistics

                    def notify() -> None:
                        with Process.audit("histogram-notify"):
                            if notify_data:
                                self.notify_property_changed("histogram_widget_data")
                            if notify_statistics:
                                self.notify_property_changed("statistics")
                            with self.__handle_lock:
                                self.__handle = None

                    if notify_data or notify_statistics:
                        with self.__handle_lock:
                            if self.__handle:
                                self.__handle.cancel()
                                self.__handle = None
                            self.__handle = self.__event_loop.call_soon_threadsafe(notify)

    # inputs

    @property
    def display_data_and_metadata(self) -> typing.Optional[DataAndMetadata.DataAndMetadata]:
        return self.__display_data_and_metadata

    @display_data_and_metadata.setter
    def display_data_and_metadata(self, value: typing.Optional[DataAndMetadata.DataAndMetadata]) -> None:
        with self.__lock:
            self.__display_data_and_metadata = value
            self.__region_data_and_metadata = None
            self.__histogram_widget_data_dirty = True
            self.__statistics_dirty = True
        self.__event.set()

    @property
    def region(self) -> typing.Optional[Graphics.Graphic]:
        return self.__region

    @region.setter
    def region(self, value: typing.Optional[Graphics.Graphic]) -> None:
        with self.__lock:
            self.__region = value
            self.__region_data_and_metadata = None
            self.__histogram_widget_data_dirty = True
            self.__statistics_dirty = True
        self.__event.set()

    @property
    def display_range(self) -> typing.Optional[typing.Tuple[float, float]]:
        return self.__display_range

    @display_range.setter
    def display_range(self, value: typing.Optional[typing.Tuple[float, float]]) -> None:
        with self.__lock:
            self.__display_range = value
            self.__histogram_widget_data_dirty = True
        self.__event.set()

    @property
    def display_data_range(self) -> typing.Optional[typing.Tuple[float, float]]:
        return self.__display_data_range

    @display_data_range.setter
    def display_data_range(self, value: typing.Optional[typing.Tuple[float, float]]) -> None:
        with self.__lock:
            self.__display_data_range = value
            self.__statistics_dirty = True
        self.__event.set()

    @property
    def displayed_intensity_calibration(self) -> typing.Optional[Calibration.Calibration]:
        return self.__displayed_intensity_calibration

    @displayed_intensity_calibration.setter
    def displayed_intensity_calibration(self, value: typing.Optional[Calibration.Calibration]) -> None:
        with self.__lock:
            self.__displayed_intensity_calibration = value
            self.__statistics_dirty = True
        self.__event.set()

    # outputs

    @property
    def histogram_widget_data(self) -> HistogramWidgetData:
        return self.__histogram_widget_data

    @histogram_widget_data.setter
    def histogram_widget_data(self, value: HistogramWidgetData) -> None:
        pass  # dummy implementation to be compatible with PropertyChangedPropertyModel

    @property
    def statistics(self) -> _StatisticsTable:
        return self.__statistics

    @statistics.setter
    def statistics(self, value: _StatisticsTable) -> None:
        pass  # dummy implementation to be compatible with PropertyChangedPropertyModel

    # private methods

    def __evaluate(self) -> None:
        try:
            with self.__lock:
                display_data_and_metadata = self.__display_data_and_metadata
                region = self.__region
                display_range = self.__display_range
                display_data_range = self.__display_data_range
                displayed_intensity_calibration = self.__displayed_intensity_calibration
                region_data_and_metadata = self.__region_data_and_metadata
                histogram_widget_data_dirty = self.__histogram_widget_data_dirty
                statistics_dirty = self.__statistics_dirty
                histogram_widget_data = self.__histogram_widget_data
                statistics = self.__statistics
                self.__histogram_widget_data_dirty = False
                self.__statistics_dirty = False
            if not region_data_and_metadata:
                region_data_and_metadata = calculate_region_data(
                    weakref.ref(display_data_and_metadata) if display_data_and_metadata else None,
                    weakref.ref(region) if region else None
                )
            if histogram_widget_data_dirty:
                histogram_widget_data = calculate_histogram_widget_data(region_data_and_metadata, display_range)
            if statistics_dirty:
                statistics = calculate_statistics(region_data_and_metadata, display_data_range, region, displayed_intensity_calibration)
            with self.__lock:
                if not self.__histogram_widget_data_dirty and not self.__statistics_dirty:
                    self.__region_data_and_metadata = region_data_and_metadata
                self.__histogram_widget_data = histogram_widget_data
                self.__statistics = statistics
        except Exception as e:
            import traceback
            traceback.print_exc()

    # test methods

    def _evaluate_immediate(self) -> None:
        self.__evaluate()
        self.notify_property_changed("histogram_widget_data")
        self.notify_property_changed("statistics")


class HistogramPanel(Panel.Panel):
    """ A panel to present a histogram of the selected data item. """

    def __init__(self, document_controller: DocumentController.DocumentController, panel_id: str,
                 properties: Persistence.PersistentDictType, debounce: bool = True, sample: bool = True) -> None:
        super().__init__(document_controller, panel_id, _("Histogram"))

        def compare_data(a: typing.Any, b: typing.Any) -> bool:
            return numpy.array_equal(a.data if a else None, b.data if b else None)  # type: ignore

        def extract_displayed_intensity_calibration(display_calibration_info: typing.Optional[DisplayItem.DisplayCalibrationInfo]) -> typing.Optional[Calibration.Calibration]:
            return display_calibration_info.displayed_intensity_calibration if display_calibration_info else None

        display_item_stream = TargetDisplayItemStream(document_controller)
        display_data_channel_stream = StreamPropertyStream[DisplayItem.DisplayDataChannel](typing.cast(Stream.AbstractStream[Observable.Observable], display_item_stream), "display_data_channel")
        display_values_stream = DisplayDataChannelDisplayValuesStream(display_data_channel_stream)
        display_data_and_metadata_stream = DisplayValuesValueStream[DataAndMetadata.DataAndMetadata](display_values_stream, "display_data_and_metadata", cmp=compare_data)
        display_range_stream = DisplayValuesValueStream[typing.Tuple[float, float]](display_values_stream, "display_range")
        display_data_range_stream = DisplayValuesValueStream[typing.Tuple[float, float]](display_values_stream, "data_range")
        displayed_intensity_calibration_stream = Stream.MapStream[DisplayItem.DisplayCalibrationInfo, Calibration.Calibration](DisplayCalibrationInfoStream(display_item_stream), extract_displayed_intensity_calibration)

        self._histogram_processor = HistogramProcessor(document_controller.event_loop)

        region_stream = TargetRegionStream(display_item_stream)
        self.__setters = [
            PropertySetter(display_data_and_metadata_stream, self._histogram_processor, "display_data_and_metadata"),
            PropertySetter(region_stream, self._histogram_processor, "region"),
            PropertySetter(display_range_stream, self._histogram_processor, "display_range"),
            PropertySetter(display_data_range_stream, self._histogram_processor, "display_data_range"),
            PropertySetter(displayed_intensity_calibration_stream, self._histogram_processor, "displayed_intensity_calibration"),
        ]

        self.__histogram_widget_data_model = Model.PropertyChangedPropertyModel[HistogramWidgetData](self._histogram_processor, "histogram_widget_data")
        self.__statistics_model = Model.PropertyChangedPropertyModel[_StatisticsTable](self._histogram_processor, "statistics")

        color_map_data_stream = StreamPropertyStream[_RGBA8ImageDataType](typing.cast(Stream.AbstractStream[Observable.Observable], display_data_channel_stream), "color_map_data", cmp=typing.cast(typing.Callable[[typing.Optional[T], typing.Optional[T]], bool], numpy.array_equal))

        def cursor_changed_fn(canvas_x: typing.Optional[float], display_range: typing.Optional[typing.Tuple[float, float]]) -> None:
            if not canvas_x:
                document_controller.cursor_changed(None)
            if display_item_stream and display_item_stream.value and canvas_x:
                if display_range is not None:  # can be None with empty data
                    displayed_intensity_calibration = display_item_stream.value.displayed_intensity_calibration
                    adjusted_x = display_range[0] + canvas_x * (display_range[1] - display_range[0])
                    adjusted_x_str = displayed_intensity_calibration.convert_to_calibrated_value_str(adjusted_x)
                    document_controller.cursor_changed([_('Intensity: ') + adjusted_x_str])
                else:
                    document_controller.cursor_changed(None)

        self.__color_map_data_model: Model.PropertyModel[_RGBA8ImageDataType] = Model.StreamValueModel(color_map_data_stream, cmp=numpy.array_equal)

        self._histogram_widget = HistogramWidget(document_controller, display_item_stream, self.__histogram_widget_data_model, self.__color_map_data_model, cursor_changed_fn)
        self._statistics_widget = StatisticsWidget(self.ui, self.__statistics_model)

        # create the main column with the histogram and the statistics section
        column = self.ui.create_column_widget(properties={"height": 80 + 18 * 3 + 12})
        column.add(self._histogram_widget)
        column.add_spacing(6)
        column.add(self._statistics_widget)
        column.add_spacing(6)
        column.add_stretch()

        # this is necessary to make the panel happy
        self.widget = column

    def close(self) -> None:
        self.__histogram_widget_data_model.close()
        self.__histogram_widget_data_model = typing.cast(typing.Any, None)
        self.__color_map_data_model.close()
        self.__color_map_data_model = typing.cast(typing.Any, None)
        self.__statistics_model.close()
        self.__statistics_model = typing.cast(typing.Any, None)
        self._statistics_widget = typing.cast(typing.Any, None)
        self.__histogram_widget_data_model = typing.cast(typing.Any, None)
        self.__color_map_data_model = typing.cast(typing.Any, None)
        self._histogram_processor.close()
        self._histogram_processor = typing.cast(typing.Any, None)
        self.__setters = typing.cast(typing.Any, None)
        super().close()


class TargetDisplayItemStream(Stream.AbstractStream[DisplayItem.DisplayItem]):

    def __init__(self, document_controller: DocumentController.DocumentController):
        super().__init__()
        # outgoing messages
        self.value_stream = Event.Event()
        # cached values
        self.__value: typing.Optional[DisplayItem.DisplayItem] = None
        # listen for selected data item changes
        self.__focused_display_item_changed_event_listener = document_controller.focused_display_item_changed_event.listen(self.__focused_display_item_changed)
        # manually send the first data item changed message to set things up.
        self.__focused_display_item_changed(document_controller.selected_display_item)

    def about_to_delete(self) -> None:
        # disconnect data item binding
        self.__value = None
        self.__focused_display_item_changed_event_listener.close()
        self.__focused_display_item_changed_event_listener = typing.cast(typing.Any, None)
        super().about_to_delete()

    @property
    def value(self) -> typing.Optional[DisplayItem.DisplayItem]:
        return self.__value

    def __focused_display_item_changed(self, display_item: typing.Optional[DisplayItem.DisplayItem]) -> None:
        if display_item != self.__value:
            self.__value = display_item
            self.value_stream.fire(display_item)


class DisplayCalibrationInfoStream(Stream.ValueStream[DisplayItem.DisplayCalibrationInfo]):
    def __init__(self, display_item_stream: TargetDisplayItemStream) -> None:
        super().__init__()
        self.__display_item_stream = display_item_stream.add_ref()
        self.__display_stream_listener = display_item_stream.value_stream.listen(ReferenceCounting.weak_partial(DisplayCalibrationInfoStream.__display_item_changed, self))
        self.__display_data_delta_stream_action: typing.Optional[Stream.ValueStreamAction[DisplayItem.DisplayDataDelta]] = None
        self.__display_item_changed(display_item_stream.value)

    def __display_item_changed(self, display_item: typing.Optional[DisplayItem.DisplayItem]) -> None:
        if display_item:
            self.__display_data_delta_stream_action = Stream.ValueStreamAction(display_item.display_data_delta_stream, ReferenceCounting.weak_partial(DisplayCalibrationInfoStream.__handle_display_data_delta_stream, self))
            self.__handle_display_data_delta_stream(display_item.display_data_delta_stream.value)
        else:
            self.__display_data_delta_stream_action = None
            self.__handle_display_data_delta_stream(None)

    def __handle_display_data_delta_stream(self, value: typing.Optional[DisplayItem.DisplayDataDelta]) -> None:
        if value:
            self.value_stream.fire(value.display_calibration_info)
        else:
            self.value_stream.fire(None)


class TargetRegionStream(Stream.AbstractStream[Graphics.Graphic]):

    def __init__(self, display_item_stream: TargetDisplayItemStream) -> None:
        super().__init__()
        # outgoing messages
        self.value_stream = Event.Event()
        # references
        self.__display_item_stream = display_item_stream.add_ref()
        # initialize
        self.__display_graphic_selection_changed_event_listener: typing.Optional[Event.EventListener] = None
        self.__value: typing.Optional[Graphics.Graphic] = None
        # listen for display changes
        self.__display_stream_listener = display_item_stream.value_stream.listen(self.__display_item_changed)
        self.__graphic_changed_event_listener: typing.Optional[Event.EventListener] = None
        self.__graphic_about_to_be_removed_event_listener: typing.Optional[Event.EventListener] = None
        self.__display_item_changed(display_item_stream.value)

    def about_to_delete(self) -> None:
        self.__display_stream_listener.close()
        self.__display_stream_listener = typing.cast(typing.Any, None)
        self.__display_item_stream.remove_ref()
        self.__display_item_stream = typing.cast(typing.Any, None)
        if self.__graphic_changed_event_listener:
            self.__graphic_changed_event_listener.close()
            self.__graphic_changed_event_listener = None
        if self.__graphic_about_to_be_removed_event_listener:
            self.__graphic_about_to_be_removed_event_listener.close()
            self.__graphic_about_to_be_removed_event_listener = None
        if self.__display_graphic_selection_changed_event_listener:
            self.__display_graphic_selection_changed_event_listener.close()
            self.__display_graphic_selection_changed_event_listener = None
        self.__value = None
        super().about_to_delete()

    @property
    def value(self) -> typing.Optional[Graphics.Graphic]:
        return self.__value

    def __display_item_changed(self, display_item: typing.Optional[DisplayItem.DisplayItem]) -> None:
        def display_graphic_selection_changed(graphic_selection: DisplayItem.GraphicSelection) -> None:
            current_index = graphic_selection.current_index
            if current_index is not None:
                assert display_item
                new_value = display_item.graphics[current_index]
                if new_value != self.__value:
                    self.__value = new_value
                    def graphic_changed(property: str) -> None:
                        self.value_stream.fire(self.__value)
                    def graphic_removed() -> None:
                        self.__value = None
                        self.value_stream.fire(None)
                    if self.__graphic_changed_event_listener:
                        self.__graphic_changed_event_listener.close()
                        self.__graphic_changed_event_listener = None
                    if self.__graphic_about_to_be_removed_event_listener:
                        self.__graphic_about_to_be_removed_event_listener.close()
                        self.__graphic_about_to_be_removed_event_listener = None
                    if self.__value:
                        self.__graphic_changed_event_listener = self.__value.property_changed_event.listen(graphic_changed)
                        self.__graphic_about_to_be_removed_event_listener = self.__value.about_to_be_removed_event.listen(graphic_removed)
                    graphic_changed("role")  # pass a dummy property
            elif self.__value is not None:
                self.__value = None
                if self.__graphic_changed_event_listener:
                    self.__graphic_changed_event_listener.close()
                    self.__graphic_changed_event_listener = None
                if self.__graphic_about_to_be_removed_event_listener:
                    self.__graphic_about_to_be_removed_event_listener.close()
                    self.__graphic_about_to_be_removed_event_listener = None
                self.value_stream.fire(None)
        if self.__graphic_changed_event_listener:
            self.__graphic_changed_event_listener.close()
            self.__graphic_changed_event_listener = None
        if self.__graphic_about_to_be_removed_event_listener:
            self.__graphic_about_to_be_removed_event_listener.close()
            self.__graphic_about_to_be_removed_event_listener = None
        if self.__display_graphic_selection_changed_event_listener:
            self.__display_graphic_selection_changed_event_listener.close()
            self.__display_graphic_selection_changed_event_listener = None
        if display_item:
            self.__display_graphic_selection_changed_event_listener = display_item.graphic_selection_changed_event.listen(display_graphic_selection_changed)
            display_graphic_selection_changed(display_item.graphic_selection)
        elif self.__value is not None:
            self.__value = None
            self.value_stream.fire(None)


class ConcatStream(Stream.AbstractStream[T], typing.Generic[T]):
    """Make a new stream for each new value of input stream and concatenate new stream output."""

    def __init__(self, stream: Stream.AbstractStream[Observable.Observable],
                 concat_fn: typing.Callable[[typing.Optional[Observable.Observable]], Stream.AbstractStream[T]]) -> None:
        super().__init__()
        # outgoing messages
        self.value_stream = Event.Event()
        # references
        self.__stream = stream.add_ref()
        # initialize
        self.__concat_fn = concat_fn
        self.__value: typing.Optional[T] = None
        self.__out_stream: typing.Optional[Stream.AbstractStream[T]] = None
        self.__out_stream_listener: typing.Optional[Event.EventListener] = None

        # define a stub and use weak_partial to avoid holding references to self.
        def stream_changed(stream: ConcatStream[T], value: typing.Optional[Observable.Observable]) -> None:
            stream.__stream_changed(value)

        self.__stream_listener = stream.value_stream.listen(ReferenceCounting.weak_partial(stream_changed, self))
        self.__stream_changed(stream.value)

    def about_to_delete(self) -> None:
        if self.__out_stream_listener:
            self.__out_stream_listener.close()
            self.__out_stream_listener = None
        if self.__out_stream:
            self.__out_stream.remove_ref()
            self.__out_stream = typing.cast(typing.Any, None)
        self.__value = None
        self.__stream_listener.close()
        self.__stream_listener = typing.cast(Event.EventListener, None)
        self.__stream.remove_ref()
        self.__stream = typing.cast(typing.Any, None)
        super().about_to_delete()

    @property
    def value(self) -> typing.Optional[T]:
        return self.__value

    def send_value(self, value: typing.Optional[T]) -> None:
        self.__value = value
        self.value_stream.fire(self.value)

    def __stream_changed(self, item: typing.Optional[Observable.Observable]) -> None:
        if self.__out_stream_listener:
            self.__out_stream_listener.close()
            self.__out_stream_listener = None
        if self.__out_stream:
            self.__out_stream.remove_ref()
            self.__out_stream = typing.cast(typing.Any, None)
        if item:
            # define a stub and use weak_partial to avoid holding references to self.
            def out_stream_changed(stream: ConcatStream[T], new_value: typing.Optional[T]) -> None:
                stream.send_value(new_value)

            self.__out_stream = self.__concat_fn(item)
            self.__out_stream.add_ref()
            self.__out_stream_listener = self.__out_stream.value_stream.listen(ReferenceCounting.weak_partial(out_stream_changed, self))
            out_stream_changed(self, self.__out_stream.value)
        else:
            self.__value = None
            self.value_stream.fire(None)


class StreamPropertyStream(ConcatStream[T], typing.Generic[T]):
    def __init__(self, stream: Stream.AbstractStream[Observable.Observable], property_name: str, cmp: typing.Optional[typing.Callable[[typing.Optional[T], typing.Optional[T]], bool]] = None) -> None:
        def fn(x: typing.Optional[Observable.Observable]) -> Stream.AbstractStream[T]:
            assert x
            return Stream.PropertyChangedEventStream[T](x, property_name, cmp)
        super().__init__(stream, fn)


class DisplayDataChannelDisplayValuesStream(Stream.ValueStream[DisplayItem.DisplayValues]):
    def __init__(self, display_data_channel_stream: Stream.AbstractStream[DisplayItem.DisplayDataChannel]) -> None:
        super().__init__()
        # initialize
        self.__display_values_subscription: typing.Optional[DisplayItem.DisplayValuesSubscription] = None
        # listen for display changes
        self.__display_data_channel_stream = display_data_channel_stream.add_ref()
        self.__display_data_channel_stream_listener = display_data_channel_stream.value_stream.listen(ReferenceCounting.weak_partial(DisplayDataChannelDisplayValuesStream.__display_data_channel_changed, self))
        self.__display_data_channel_changed(display_data_channel_stream.value)

    def about_to_delete(self) -> None:
        self.__display_values_subscription = None
        self.__display_data_channel_stream_listener.close()
        self.__display_data_channel_stream_listener = typing.cast(typing.Any, None)
        self.__display_data_channel_stream.remove_ref()
        self.__display_data_channel_stream = typing.cast(typing.Any, None)
        super().about_to_delete()

    def __display_values_changed(self, display_values: typing.Optional[DisplayItem.DisplayValues]) -> None:
        self.send_value(display_values)

    def __display_data_channel_changed(self, display_data_channel: typing.Optional[DisplayItem.DisplayDataChannel]) -> None:
        if display_data_channel:
            self.__display_values_subscription = display_data_channel.subscribe_to_latest_computed_display_values(ReferenceCounting.weak_partial(DisplayDataChannelDisplayValuesStream.__display_values_changed, self))
            self.__display_values_changed(display_data_channel.get_latest_computed_display_values())
        else:
            self.__display_values_subscription = None
            self.send_value(None)


class DisplayValuesValueStream(Stream.ValueStream[T]):
    def __init__(self, display_value_stream: Stream.AbstractStream[DisplayItem.DisplayValues], property_name: str, cmp: typing.Optional[typing.Callable[[typing.Optional[T], typing.Optional[T]], bool]] = None) -> None:
        super().__init__()
        self.__property_name = property_name
        self.__cmp: typing.Callable[[typing.Optional[T], typing.Optional[T]], bool] = cmp if cmp else typing.cast(typing.Callable[[typing.Optional[T], typing.Optional[T]], bool], operator.eq)
        # listen for display changes
        self.__display_value_stream = display_value_stream
        self.__display_value_stream_listener = display_value_stream.value_stream.listen(ReferenceCounting.weak_partial(DisplayValuesValueStream.__display_values_changed, self))
        self.__display_values_changed(display_value_stream.value)

    def about_to_delete(self) -> None:
        self.__display_value_stream_listener.close()
        self.__display_value_stream_listener = typing.cast(typing.Any, None)
        self.__display_value_stream = typing.cast(typing.Any, None)
        super().about_to_delete()

    def __display_values_changed(self, display_values: typing.Optional[DisplayItem.DisplayValues]) -> None:
        new_value = getattr(display_values, self.__property_name) if display_values else None
        if not self.__cmp(new_value, self.value):
            self.send_value(new_value)



from __future__ import annotations

# standard libraries
# None

# third party libraries
# None

# local libraries
from nion.utils import Event
from nion.utils import Observable


class Activity(Observable.Observable):
    def __init__(self, activity_id: str, title: str) -> None:
        super().__init__()
        self.activity_id = activity_id
        self.title = title

    @property
    def displayed_title(self) -> str:
        return self.title


activity_appended_event = Event.Event()
activity_finished_event = Event.Event()


def append_activity(activity: Activity) -> None:
    # in order to be able to be called on other threads, just send an event. the activity panel will listen.
    activity_appended_event.fire(activity)


def activity_finished(activity: Activity) -> None:
    # in order to be able to be called on other threads, just send an event. the activity panel will listen.
    activity_finished_event.fire(activity)
